#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import argparse


def fix_line(line, site, filetype):
    """
    Squashes the entries for multiple sites into one.
    This is required when entries are defined for a different
    site than they are reported.
    Such situation happend e.g.
    for BRAM_[LR]. All the entries are defined for RAMBFIFO36E1,
    while they are reported for RAMB18E1 or FIFO18E1

    Parameters
    ----------
    line: str
        raw dump file line
    site: str
        site to which all the entries will be squashed
    filetype: str
        entries type. One of [timings, pins, properties]

    Returns
    -------
    str
        line with squashed entries
    """

    assert filetype in [
        "timings", "pins", "properties"
    ], "Unsupported filetype"

    line = line.split()
    newline = list()
    sites_count = int(line[1])
    newline.append(line[0])
    # we'll emit only one site
    newline.append("1")
    newline.append(site)
    newline.append("1")
    newline.append(site)
    entries = list()
    all_entries = 0
    loc = 2
    for site in range(0, sites_count):
        bels_count = int(line[loc + 1])
        loc += 2
        for bel in range(0, bels_count):
            entries_count = int(line[loc + 1])
            loc += 2
            all_entries += entries_count
            for entry in range(0, entries_count):
                if filetype == 'timings':
                    for delay_word in range(0, 6):
                        entries.append(line[loc])
                        loc += 1
                elif filetype == 'pins':
                    for pin_word in range(0, 4):
                        entries.append(line[loc])
                        loc += 1
                elif filetype == 'properties':
                    entries.append(line[loc])
                    loc += 1
                    values_count = int(line[loc])
                    entries.append(line[loc])
                    loc += 1
                    for value in range(0, values_count):
                        entries.append(line[loc])
                        loc += 1
    newline.append(str(all_entries))
    newline.extend(entries)
    return " ".join(newline) + "\n"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--txtin', type=str, help='Input text file')
    parser.add_argument('--txtout', type=str, help='Output text file')
    parser.add_argument(
        '--site',
        type=str,
        help='Site to which the entries should be squashed')
    parser.add_argument(
        '--slice', type=str, help='Slice for which the entries shall be fixed')
    parser.add_argument(
        '--type', type=str, choices=['timings', 'pins', 'properties'])

    args = parser.parse_args()
    lines = list()

    with open(args.txtin, 'r') as fp:
        for line in fp:
            if line.startswith(args.slice):
                line = fix_line(line, args.site, args.type)
            lines.append(line)

    with open(args.txtout, 'w') as fp:
        for line in lines:
            fp.write(line)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
class LutMaker(object):
    def __init__(self):
        self.input_lut_idx = 0
        self.output_lut_idx = 0
        self.lut_input_idx = 0

    def get_next_input_net(self):
        net = 'lut_{}_i[{}]'.format(self.input_lut_idx, self.lut_input_idx)

        self.lut_input_idx += 1
        if self.lut_input_idx > 5:
            self.lut_input_idx = 0
            self.input_lut_idx += 1

        return net

    def get_next_output_net(self):
        net = 'lut_{}_o'.format(self.output_lut_idx)
        self.output_lut_idx += 1
        return net

    def create_wires_and_luts(self):
        if self.output_lut_idx > self.input_lut_idx:
            nluts = self.output_lut_idx
        else:
            nluts = self.input_lut_idx
            if self.lut_input_idx > 0:
                nluts += 1

        for lut in range(nluts):
            yield """
            wire [5:0] lut_{lut}_i;
            wire lut_{lut}_o;

            (* KEEP, DONT_TOUCH *)
            LUT6 lut_{lut} (
                .I0(lut_{lut}_i[0]),
                .I1(lut_{lut}_i[1]),
                .I2(lut_{lut}_i[2]),
                .I3(lut_{lut}_i[3]),
                .I4(lut_{lut}_i[4]),
                .I5(lut_{lut}_i[5]),
                .O(lut_{lut}_o)
            );
            """.format(lut=lut)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Spartan 6 bitstream analyzer tool.

This script reads a Spartan6 bitstream and prints out some useful information.
It can also create a frames file with the configuration data words.
The bitstream is analyzed word by word and interpreted according to
the UG380 Configuration User Guide.

The tool can be used to derive the initialization, startup and finalization
sequence as well as the configuration data. The latter is written to a frames
file which can be used by the bitstream tools such as frames2bit to generate
a valid bitstream.
'''

import argparse
from io import StringIO

from prjxray.util import OpenSafeFile

conf_regs = {
    0: "CRC",
    1: "FAR_MAJ",
    2: "FAR_MIN",
    3: "FDRI",
    4: "FDRO",
    5: "CMD",
    6: "CTL",
    7: "MASK",
    8: "STAT",
    9: "LOUT",
    10: "COR1",
    11: "COR2",
    12: "PWRDN_REG",
    13: "FLR",
    14: "IDCODE",
    15: "CWDT",
    16: "HC_OPT_REG",
    18: "CSBO",
    19: "GENERAL1",
    20: "GENERAL2",
    21: "GENERAL3",
    22: "GENERAL4",
    23: "GENERAL5",
    24: "MODE_REG",
    25: "PU_GWE",
    26: "PU_GTS",
    27: "MFWR",
    28: "CCLK_FREQ",
    29: "SEU_OPT",
    30: "EXP_SIGN",
    31: "RDBK_SIGN",
    32: "BOOTSTS",
    33: "EYE_MASK",
    34: "CBC_REG"
}

cmd_reg_codes = {
    0: "NULL",
    1: "WCFG",
    2: "MFW",
    3: "LFRM",
    4: "RCFG",
    5: "START",
    7: "RCRC",
    8: "AGHIGH",
    10: "GRESTORE",
    11: "SHUTDOWN",
    13: "DESYNC",
    14: "IPROG"
}

opcodes = ("NOP", "READ", "WRITE", "UNKNOWN")


def KnuthMorrisPratt(text, pattern):
    '''
    Yields all starting positions of copies of the pattern in the text.
    Calling conventions are similar to string.find, but its arguments can be
    lists or iterators, not just strings, it returns all matches, not just
    the first one, and it does not need the whole text in memory at once.
    Whenever it yields, it will have read the text exactly up to and including
    the match that caused the yield.
    '''

    # allow indexing into pattern and protect against change during yield
    pattern = list(pattern)

    # build table of shift amounts
    shifts = [1] * (len(pattern) + 1)
    shift = 1
    for pos in range(len(pattern)):
        while shift <= pos and pattern[pos] != pattern[pos - shift]:
            shift += shifts[pos - shift]
        shifts[pos + 1] = shift

    # do the actual search
    startPos = 0
    matchLen = 0
    for c in text:
        while matchLen == len(pattern) or \
              matchLen >= 0 and pattern[matchLen] != c:
            startPos += shifts[matchLen]
            matchLen -= shifts[matchLen]
        matchLen += 1
        if matchLen == len(pattern):
            yield startPos


class Bitstream:
    def __init__(self, file_name, verbose=False):
        self.frame_data = []
        self.idcode = 0
        self.exp_sign = 0
        self.far_min = 0
        self.far_maj = 0
        self.curr_fdri_write_len = 0
        self.curr_crc_check = 0
        self.fdri_in_progress = False
        with OpenSafeFile(file_name, "rb") as f:
            self.bytes = f.read()
        pos, self.header = self.get_header()
        self.body = [
            (i << 8) | j
            for i, j in zip(self.bytes[pos::2], self.bytes[pos + 1::2])
        ]
        self.parse_bitstream(verbose)

    def get_header(self):
        pos = next(KnuthMorrisPratt(self.bytes, [0xaa, 0x99, 0x55, 0x66]))
        return pos + 4, self.bytes[:pos + 4]

    def parse_bitstream(self, verbose):
        payload_len = 0
        for word in self.body:
            if payload_len > 0:
                if verbose:
                    print("\tWord: ", hex(word))
                payload_len = self.parse_reg(
                    reg_addr, word, payload_len, verbose)
                continue
            else:
                packet_header = self.parse_packet_header(word)
                opcode = packet_header["opcode"]
                reg_addr = packet_header["reg_addr"]
                words = packet_header["word_count"]
                type = packet_header["type"]
                if verbose:
                    print(
                        "\tWord: ", hex(word),
                        'Type: {}, Op: {}, Addr: {}, Words: {}'.format(
                            type, opcodes[opcode], reg_addr, words))
                if opcode and reg_addr in conf_regs:
                    payload_len = words
                    continue

    def parse_packet_header(self, word):
        type = (word >> 13) & 0x7
        opcode = (word >> 11) & 0x3
        reg_addr = (word >> 5) & 0x3F
        if type == 1:
            word_count = word & 0x1F
        elif type == 2:
            word_count = 2
        else:
            word_count = 0
        return {
            "type": type,
            "opcode": opcode,
            "reg_addr": reg_addr,
            "word_count": word_count
        }

    def parse_command(self, word):
        return cmd_reg_codes[word]

    def parse_cor1(self, word):
        return word

    def parse_cor2(self, word):
        return word

    def parse_ctl(self, word):
        #decryption
        dec = (word >> 6) & 1
        #security bits
        sb = (word >> 4) & 3
        #persist
        p = (word >> 3) & 1
        #use efuse
        efuse = (word >> 2) & 1
        #crc extstat disable
        crc = (word >> 1) & 1
        return {
            "decryption": dec,
            "security bits": sb,
            "pesist": p,
            "use efuse": efuse,
            "crc extstat disable": crc
        }

    def parse_cclk_freq(self, word):
        ext_mclk = (word >> 14) & 1
        mclk_freq = word & 0x3FF
        return (ext_mclk, mclk_freq)

    def parse_pwrdn(self, word):
        en_eyes = (word >> 14) & 1
        filter_b = (word >> 5) & 1
        en_pgsr = (word >> 4) & 1
        en_pwrdn = (word >> 2) & 1
        keep_sclk = word & 1
        return {
            "en_eyes": en_eyes,
            "filter_b": filter_b,
            "en_pgsr": en_pgsr,
            "en_pwrdn": en_pwrdn,
            "keep_sclk": keep_sclk
        }

    def parse_eye_mask(self, word):
        return word & 0xFF

    def parse_hc_opt(self, word):
        return (word >> 6) & 1

    def parse_cwdt(self, word):
        return word

    def parse_pu_gwe(self, word):
        return word & 0x3FF

    def parse_pu_gts(self, word):
        return word & 0x3FF

    def parse_mode(self, word):
        new_mode = (word >> 13) & 0x1
        buswidth = (word >> 11) & 0x3
        bootmode = (word >> 8) & 0x7
        bootvsel = word & 0xFF
        return {
            "new_mode": new_mode,
            "buswidth": buswidth,
            "bootmode": bootmode,
            "bootvsel": bootvsel
        }

    def parse_seu(self, word):
        seu_freq = (word >> 4) & 0x3FF
        seu_run_on_err = (word >> 3) & 0x1
        glut_mask = (word >> 1) & 0x1
        seu_enable = word & 0x1
        return {
            "seu_freq": seu_freq,
            "seu_run_on_err": seu_run_on_err,
            "glut_mask": glut_mask,
            "seu_enable": seu_enable
        }

    def parse_reg(self, reg_addr, word, payload_len, verbose):
        reg = conf_regs[reg_addr]
        if reg == "CMD":
            command = self.parse_command(word)
            if verbose:
                print("Command: {}\n".format(command))
        elif reg == "FLR":
            frame_length = word
            if verbose:
                print("Frame length: {}\n".format(frame_length))
        elif reg == "COR1":
            conf_options = self.parse_cor1(word)
            if verbose:
                print("COR1 options: {}\n".format(conf_options))
        elif reg == "COR2":
            conf_options = self.parse_cor2(word)
            if verbose:
                print("COR2 options: {}\n".format(conf_options))
        elif reg == "IDCODE":
            assert payload_len < 3
            if payload_len == 2:
                self.idcode = word << 16
            elif payload_len == 1:
                self.idcode |= word
                if verbose:
                    print("IDCODE: {}\n".format(hex(self.idcode)))
        elif reg == "MASK":
            mask = word
            if verbose:
                print("Mask value: {}\n".format(mask))
        elif reg == "CTL":
            ctl_options = self.parse_ctl(word)
            if verbose:
                print("CTL options: {}\n".format(ctl_options))
        elif reg == "CCLK_FREQ":
            cclk_freq_options = self.parse_cclk_freq(word)
            if verbose:
                print("CCLK_FREQ options: {}\n".format(cclk_freq_options))
        elif reg == "PWRDN_REG":
            suspend_reg_options = self.parse_pwrdn(word)
            if verbose:
                print("{} options: {}\n".format(reg, suspend_reg_options))
        elif reg == "EYE_MASK":
            eye_mask = self.parse_eye_mask(word)
            if verbose:
                print("{} options: {}\n".format(reg, eye_mask))
        elif reg == "HC_OPT_REG":
            hc_options = self.parse_hc_opt(word)
            if verbose:
                print("{} options: {}\n".format(reg, hc_options))
        elif reg == "CWDT":
            cwdt_options = self.parse_cwdt(word)
            if verbose:
                print("{} options: {}\n".format(reg, cwdt_options))
        elif reg == "PU_GWE":
            pu_gwe_sequence = self.parse_pu_gwe(word)
            if verbose:
                print("{} options: {}\n".format(reg, pu_gwe_sequence))
        elif reg == "PU_GTS":
            pu_gts_sequence = self.parse_pu_gts(word)
            if verbose:
                print("{} options: {}\n".format(reg, pu_gts_sequence))
        elif reg == "MODE_REG":
            mode_options = self.parse_mode(word)
            if verbose:
                print("{} options: {}\n".format(reg, mode_options))
        elif reg == "GENERAL1" or reg == "GENERAL2" \
             or reg == "GENERAL3" or reg == "GENERAL4" \
             or reg == "GENERAL5":
            general_options = word
            if verbose:
                print("{} options: {}\n".format(reg, general_options))
        elif reg == "SEU_OPT":
            seu_options = self.parse_seu(word)
            if verbose:
                print("{} options: {}\n".format(reg, seu_options))
        elif reg == "EXP_SIGN":
            if payload_len == 2:
                self.exp_sign = word << 16
            elif payload_len == 1:
                self.exp_sign |= word
                if verbose:
                    print("{}: {}\n".format(reg, self.exp_sign))
        elif reg == "FAR_MAJ":
            if payload_len == 2:
                self.current_far_maj = word
            elif payload_len == 1:
                self.current_far_min = word
                if verbose:
                    print(
                        "{}: {} FAR_MIN: {}\n".format(
                            reg, self.far_maj, self.far_min))
        elif reg == "FDRI":
            if self.fdri_in_progress:
                self.frame_data.append(word)
                if payload_len == 1:
                    self.fdri_in_progress = False
                    return 0
            elif payload_len == 2:
                self.curr_fdri_write_len = (word & 0xFFF) << 16
            elif payload_len == 1:
                self.curr_fdri_write_len |= word
                self.fdri_in_progress = True
                # Check if 0 words actually means read something
                payload_len = self.curr_fdri_write_len + 2
                if verbose:
                    print("{}: {}\n".format(reg, self.curr_fdri_write_len))
                return payload_len
        elif reg == "CRC":
            if payload_len == 2:
                self.curr_crc_check = (word & 0xFFF) << 16
            elif payload_len == 1:
                self.curr_crc_check |= word
                if verbose:
                    print("{}: {}\n".format(reg, self.curr_crc_check))
        payload_len -= 1
        return payload_len

    def write_frames_txt(self, file_name):
        '''Write frame data in a more readable format'''
        frame_stream = StringIO()
        for i in range(len(self.frame_data)):
            if i % 65 == 0:
                frame_stream.write("\nFrame {:4}\n".format(i // 65))
            #IOB word
            if i % 65 == 32:
                frame_stream.write(
                    "\n#{:3}:{:6}\n".format(i % 65, hex(self.frame_data[i])))
            else:
                frame_stream.write(
                    "#{:3}:{:6},".format(i % 65, hex(self.frame_data[i])))
        with OpenSafeFile(file_name, "w") as f:
            print(frame_stream.getvalue(), file=f)

    def write_frames(self, file_name):
        '''Write configuration data to frames file'''
        frame_stream = StringIO()
        for i in range(len(self.frame_data)):
            if i % 65 == 0:
                frame_stream.write("0x{:08x} ".format(i // 65))
            frame_stream.write("0x{:04x}".format(self.frame_data[i]))
            if i % 65 == 64:
                frame_stream.write("\n")
            elif i < len(self.frame_data) - 1:
                frame_stream.write(",")
        with OpenSafeFile(file_name, "w") as f:
            print(frame_stream.getvalue(), file=f)


def main(args):
    verbose = not args.silent
    bitstream = Bitstream(args.bitstream, verbose)
    print("Frame data length: ", len(bitstream.frame_data))
    if args.frames_out:
        bitstream.write_frames(args.frames_out)
        if verbose:
            bitstream.write_frames_txt(args.frames_out + ".txt")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--bitstream', help='Input bitstream')
    parser.add_argument('--frames_out', help='Output frames file')
    parser.add_argument(
        '--silent', help="Don't print analysis details", action='store_true')
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker, add_site_group_zero
from prjxray.verilog import to_int
from prjxray.verilog import quote
import json


def add(segmk, site, dsp, tag, bit, value, invert, is_vector=True):
    tag = "%s.%s%s%s" % (
        dsp, ('Z' if invert else ''), tag, '[%u]' % bit if is_vector else '')
    value = (~value if invert else value)
    value >>= bit
    return segmk.add_site_tag(site, tag, value & 1)


def run():
    segmk = Segmaker("design.bits", verbose=True)

    print("Loading tags")
    with open('params.json', 'r') as fp:
        data = json.load(fp)

    used_dsps = set()

    for params in data['instances']:
        dsp = "DSP_0" if params['SITE'][-1] in "02468" else "DSP_1"
        site = params['SITE']

        if params['USE_DPORT'] == quote(
                "TRUE") and params['USE_MULT'] != quote("NONE"):
            add(segmk, site, dsp, 'ADREG', 0, to_int(params['ADREG']), 1)
        add(segmk, site, dsp, 'ALUMODEREG', 0, to_int(params['ALUMODEREG']), 1)

        if params['A_INPUT'] == quote("DIRECT"):
            add(
                segmk, site, dsp, 'AREG_0', 0,
                int(to_int(params['AREG']) == 0), 0, False)
            add(
                segmk, site, dsp, 'AREG_2', 0,
                int(to_int(params['AREG']) == 2), 0, False)

        if params['B_INPUT'] == quote("DIRECT"):
            add(
                segmk, site, dsp, 'BREG_0', 0,
                int(to_int(params['BREG']) == 0), 0, False)
            add(
                segmk, site, dsp, 'BREG_2', 0,
                int(to_int(params['BREG']) == 2), 0, False)

        if params['A_INPUT'] == quote("CASCADE"):
            add(
                segmk, site, dsp, 'AREG_2_ACASCREG_1', 0,
                int(
                    to_int(params['AREG']) == 2
                    and to_int(params['ACASCREG']) == 1), 0, False)
            add(
                segmk, site, dsp, 'AREG_2_ACASCREG_1', 0,
                int(
                    to_int(params['AREG']) == 2
                    and to_int(params['ACASCREG']) == 1), 1, False)

        if params['B_INPUT'] == quote("CASCADE"):
            add(
                segmk, site, dsp, 'BREG_2_BCASCREG_1', 0,
                int(
                    to_int(params['BREG']) == 2
                    and to_int(params['BCASCREG']) == 1), 0, False)
            add(
                segmk, site, dsp, 'BREG_2_BCASCREG_1', 0,
                int(
                    to_int(params['BREG']) == 2
                    and to_int(params['BCASCREG']) == 1), 1, False)

        add(segmk, site, dsp, 'CARRYINREG', 0, to_int(params['CARRYINREG']), 1)
        add(
            segmk, site, dsp, 'CARRYINSELREG', 0,
            to_int(params['CARRYINSELREG']), 1)
        add(segmk, site, dsp, 'CREG', 0, to_int(params['CREG']), 1)
        if params['USE_DPORT'] == quote(
                "TRUE") and params['USE_MULT'] != quote("NONE"):
            add(segmk, site, dsp, 'DREG', 0, to_int(params['DREG']), 1)
        add(segmk, site, dsp, 'INMODEREG', 0, to_int(params['INMODEREG']), 1)
        add(segmk, site, dsp, 'OPMODEREG', 0, to_int(params['OPMODEREG']), 1)
        add(segmk, site, dsp, 'PREG', 0, to_int(params['PREG']), 1)

        INPUT = {}
        INPUT[quote('DIRECT')] = 0
        INPUT[quote('CASCADE')] = 1

        add(segmk, site, dsp, 'A_INPUT', 0, INPUT[params['A_INPUT']], 0)
        add(segmk, site, dsp, 'B_INPUT', 0, INPUT[params['B_INPUT']], 0)

        BOOL = {}
        BOOL[quote('FALSE')] = 0
        BOOL[quote('TRUE')] = 1

        add(segmk, site, dsp, 'USE_DPORT', 0, BOOL[params['USE_DPORT']], 0)

        add(
            segmk, site, dsp, 'USE_SIMD_FOUR12', 0,
            params['USE_SIMD'] == quote("FOUR12"), 0, False)
        add(
            segmk, site, dsp, 'USE_SIMD_FOUR12_TWO24', 0,
            params['USE_SIMD'] in (quote("TWO24"), quote("FOUR12")), 0, False)

        MULT = {}
        MULT[quote('NONE')] = 0
        MULT[quote('MULTIPLY')] = 1
        MULT[quote('DYNAMIC')] = 2

        for i in range(2):
            add(segmk, site, dsp, 'USE_MULT', i, MULT[params['USE_MULT']], 0)

        add(segmk, site, dsp, 'MREG', 0, to_int(params['MREG']), 1)

        AUTORESET = {}
        AUTORESET[quote('NO_RESET')] = 0
        AUTORESET[quote('RESET_NOT_MATCH')] = 1
        AUTORESET[quote('RESET_MATCH')] = 2

        add(
            segmk, site, dsp, 'AUTORESET_PATDET_RESET_NOT_MATCH', 0,
            params['AUTORESET_PATDET'] == quote("RESET_NOT_MATCH"), 0, False)
        add(
            segmk, site, dsp, 'AUTORESET_PATDET_RESET', 0,
            params['AUTORESET_PATDET'] in (
                quote("RESET_NOT_MATCH"), quote("RESET_MATCH")), 0, False)

        for i in range(48):
            add(segmk, site, dsp, 'MASK', i, to_int(params['MASK']), 0)

        for i in range(48):
            add(segmk, site, dsp, 'PATTERN', i, to_int(params['PATTERN']), 0)

        if params['USE_PATTERN_DETECT'] == quote("PATDET"):
            add_site_group_zero(
                segmk, site, dsp + ".", [
                    "SEL_MASK_%s" % x
                    for x in ["MASK", "C", "ROUNDING_MODE1", "ROUNDING_MODE2"]
                ], "SEL_MASK_MASK", "SEL_MASK_%s" % (params['SEL_MASK'][1:-1]))

        USE_PATTERN_DETECT = {}
        USE_PATTERN_DETECT[quote('NO_PATDET')] = 0
        USE_PATTERN_DETECT[quote('PATDET')] = 1

        add(
            segmk, site, dsp, 'USE_PATTERN_DETECT', 0,
            USE_PATTERN_DETECT[params['USE_PATTERN_DETECT']], 0)

        inv_ports = [
            ("ALUMODE", 4),
            ("CARRYIN", 1),
            ("CLK", 1),
            ("INMODE", 5),
            ("OPMODE", 7),
        ]

        for port, width in inv_ports:
            param = 'IS_{}_INVERTED'.format(port)
            for i in range(width):
                add(
                    segmk, site, dsp, param, i, to_int(params[param]), 1,
                    width > 1)

    segmk.compile()
    segmk.write()


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os.path
import csv
import pickle
import re
from collections import namedtuple

from prjxray.util import OpenSafeFile


def read_root_csv(root_dir):
    """ Reads root.csv from raw db directory.

  This should only be used during database generation.

  """
    tiles = {}
    nodes = []

    with OpenSafeFile(os.path.join(root_dir, 'root.csv')) as f:
        for d in csv.DictReader(f):
            if d['filetype'] == 'tile':
                if d['subtype'] not in tiles:
                    tiles[d['subtype']] = []

                tiles[d['subtype']].append(
                    os.path.join(root_dir, d['filename']))
            elif d['filetype'] == 'node':
                nodes.append(os.path.join(root_dir, d['filename']))

    return tiles, nodes


def verify_nodes(raw_nodes, nodes, error_nodes):
    """ Compares raw_nodes with generated_nodes and adds errors to error_nodes.

  Args:
    raw_nodes - Iterable of (node name, iterable of wires in node).
    nodes - Iterable of iterable of wires in nodes.
    error_nodes - List to be appended to when an error occurs.  Elements will
                  be 3 tuple of raw node name, raw node, and generated node
                  that did not match.

  """
    wire_nodes = {}
    for node in nodes:
        node_set = set(node)
        for wire in node:
            wire_nodes[wire] = node_set

    for node, raw_node_wires in raw_nodes:
        raw_node_set = set(raw_node_wires)

        for wire in sorted(raw_node_set):
            if wire not in wire_nodes:
                if set((wire, )) != raw_node_set:
                    error_nodes.append((node, tuple(raw_node_set), (wire, )))
            elif wire_nodes[wire] != raw_node_set:
                error_nodes.append(
                    (node, tuple(raw_node_set), tuple(wire_nodes[wire])))


def check_errors(flat_error_nodes, ignored_wires):
    """ Check if error_nodes has errors that are not covered in ignored_wires.

  Args:
    flat_error_nodes - List of error_nodes generated from verify_nodes.
    ignored_wires - List of wires that should be ignored if they were generated.

  """

    error_nodes = {}
    for node, raw_node, generated_nodes in flat_error_nodes:
        if node not in error_nodes:
            error_nodes[node] = {
                'raw_node': set(raw_node),
                'generated_nodes': set(),
            }

        # Make sure all raw nodes are the same.
        assert error_nodes[node]['raw_node'] == set(raw_node)

        error_nodes[node]['generated_nodes'].add(
            tuple(sorted(generated_nodes)))

    for node, error in error_nodes.items():
        combined_generated_nodes = set()
        for generated_node in error['generated_nodes']:
            combined_generated_nodes |= set(generated_node)

        # Make sure there are not extra wires in nodes.
        assert error['raw_node'] == combined_generated_nodes, (node, error)

        good_node = max(error['generated_nodes'], key=lambda x: len(x))
        bad_nodes = error['generated_nodes'] - set((good_node, ))

        # Max sure only single wires are stranded
        assert max(len(generated_node) for generated_node in bad_nodes) == 1

        for generate_node in bad_nodes:
            for wire in generate_node:
                if wire not in ignored_wires:
                    return False

    return True


class NodeLookup(object):
    def __init__(self):
        self.nodes = {}

    def load_from_nodes(self, nodes):
        self.nodes = nodes

    def load_from_root_csv(self, nodes):
        import pyjson5 as json5
        import progressbar
        for node in progressbar.progressbar(nodes):
            with OpenSafeFile(node) as f:
                node_wires = json5.load(f)
                assert node_wires['node'] not in self.nodes
                self.nodes[node_wires['node']] = node_wires['wires']

    def load_from_file(self, fname):
        with OpenSafeFile(fname, 'rb') as f:
            self.nodes = pickle.load(f)

    def save_to_file(self, fname):
        with OpenSafeFile(fname, 'wb') as f:
            pickle.dump(self.nodes, f)

    def site_pin_node_to_wires(self, tile, node):
        if node is None:
            return

        node_wires = self.nodes[node]

        for wire in node_wires:
            if wire['wire'].startswith(tile + '/'):
                yield wire['wire'][len(tile) + 1:]

    def wires_for_tile(self, tile):
        for node in self.nodes.values():
            for wire in node:
                if wire['wire'].startswith(tile + '/'):
                    yield wire['wire'][len(tile) + 1:]


def compare_prototype_site(proto_a, proto_b):
    """ Compare two proto site type.

  Will assert if prototypes are not equivalent.

  """
    assert proto_a == proto_b, repr((proto_a, proto_b))


# All site names appear to follow the pattern <type>_X<abs coord>Y<abs coord>.
# Generally speaking, only the tile relatively coordinates are required to
# assemble arch defs, so we re-origin the coordinates to be relative to the tile
# (e.g. start at X0Y0) and discard the prefix from the name.
SITE_COORDINATE_PATTERN = re.compile('^(.+)_X([0-9]+)Y([0-9]+)$')

SiteCoordinate = namedtuple('SiteCoordinate', 'prefix x_coord y_coord')


def get_site_coordinate_from_name(name):
    """
    >>> get_site_coordinate_from_name('SLICE_X1Y0')
    SiteCoordinate(prefix='SLICE', x_coord=1, y_coord=0)

    >>> get_site_coordinate_from_name('SLICE_X0Y0')
    SiteCoordinate(prefix='SLICE', x_coord=0, y_coord=0)

    >>> get_site_coordinate_from_name('INT_L_X500Y999')
    SiteCoordinate(prefix='INT_L', x_coord=500, y_coord=999)

    """
    coordinate = SITE_COORDINATE_PATTERN.match(name)
    assert coordinate is not None, name

    return SiteCoordinate(
        prefix=coordinate.group(1),
        x_coord=int(coordinate.group(2)),
        y_coord=int(coordinate.group(3)),
    )


def get_site_prefix_from_name(name):
    """
    Returns the prefix of a site from its name
    """
    coordinate = SITE_COORDINATE_PATTERN.match(name)
    assert coordinate is not None, name

    return coordinate.group(1)


def find_origin_coordinate(site_name, site_names):
    """
    Find the coordinates of each site within the tile, and then subtract the
    smallest coordinate to re-origin them all to be relative to the tile.
    """

    x_coords = []
    y_coords = []
    for site in site_names:
        coordinate = get_site_coordinate_from_name(site)
        site_name_prefix = get_site_prefix_from_name(site_name)
        if coordinate.prefix == site_name_prefix:
            x_coords.append(coordinate.x_coord)
            y_coords.append(coordinate.y_coord)

    if len(x_coords) == 0 or len(y_coords) == 0:
        return 0, 0

    min_x_coord = min(x_coords)
    min_y_coord = min(y_coords)

    return min_x_coord, min_y_coord



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import random
random.seed(0)
import os
import re
from prjxray import verilog
from prjxray import util

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 400 + int(INCREMENT)
SLICEX, SLICEY = util.site_xy_minmax([
    'SLICEL',
    'SLICEM',
])
# 800
SLICEN = (SLICEY[1] - SLICEY[0]) * (SLICEX[1] - SLICEX[0])
print('//SLICEX: %s' % str(SLICEX))
print('//SLICEY: %s' % str(SLICEY))
print('//SLICEN: %s' % str(SLICEN))
print('//Requested CLBs: %s' % str(CLBN))


# Rearranged to sweep Y so that carry logic is easy to allocate
# XXX: careful...if odd number of Y in ROI will break carry
def gen_slices():
    for slicex in range(*SLICEX):
        for slicey in range(*SLICEY):
            # caller may reject position if needs more room
            yield ("SLICE_X%dY%d" % (slicex, slicey), (slicex, slicey))


DIN_N = CLBN * 8
DOUT_N = CLBN * 8

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.csv', 'w')
f.write('module,loc,loc2\n')
slices = gen_slices()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
for i in range(CLBN):
    # Don't have an O6 example
    modules = ['clb_PRECYINIT_' + x for x in ['0', '1', 'AX', 'CIN']]
    loc, loc_pos = next(slices)
    while True:
        module = random.choice(modules)

        if module == 'clb_PRECYINIT_CIN':
            # Need at least extra Y for CIN extra CLB
            if loc_pos[1] >= SLICEY[1] - 1:
                continue
            loc_co = loc
            loc_ci, _pos_ci = next(slices)
            params = '.LOC_CO("%s"), .LOC_CI("%s")' % (loc_co, loc_ci)
            # Don't really care about co, but add for completeness
            paramsc = loc_ci + ',' + loc_co
        else:
            params = '.LOC("%s")' % loc
            paramsc = loc + ',' + ''
        break

    print('    %s' % module)
    print('            #(%s)' % (params))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 8]), .dout(dout[  %d +: 8]));'
        % (i, 8 * i, 8 * i))

    f.write('%s,%s\n' % (module, paramsc))
f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module clb_PRECYINIT_0 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";

	(* LOC=LOC, KEEP, DONT_TOUCH *)
    CARRY4 carry4(.O(), .CO(), .DI(din[3:0]), .S(din[7:4]), .CYINIT(1'b0), .CI());
endmodule

module clb_PRECYINIT_1 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";

	(* LOC=LOC, KEEP, DONT_TOUCH *)
    CARRY4 carry4(.O(), .CO(), .DI(din[3:0]), .S(din[7:4]), .CYINIT(1'b1), .CI());
endmodule

module clb_PRECYINIT_AX (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";

	(* LOC=LOC, KEEP, DONT_TOUCH *)
    CARRY4 carry4(.O(), .CO(), .DI(din[3:0]), .S(din[7:4]), .CYINIT(din[0]), .CI());
endmodule

module clb_PRECYINIT_CIN (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC_CO="SLICE_FIXME";
    parameter LOC_CI="SLICE_FIXME";

    wire [3:0] co;

    //Gets CI
	(* LOC=LOC_CI, KEEP, DONT_TOUCH *)
    CARRY4 carry4_co(.O(), .CO(), .DI(din[3:0]), .S(din[7:4]), .CYINIT(co[3]), .CI());
    //Sends CO
	(* LOC=LOC_CO, KEEP, DONT_TOUCH *)
    CARRY4 carry4_ci(.O(), .CO(co), .DI(din[3:0]), .S(din[7:4]), .CYINIT(1'b0), .CI());
endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os, random
random.seed(0)
from prjxray import util
from prjxray import verilog

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 400 + int(INCREMENT)
print('//Requested CLBs: %s' % str(CLBN))


def gen_slices():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites(
        ['SLICEL', 'SLICEM']):
        yield site_name


DIN_N = CLBN * 8
DOUT_N = CLBN * 8

lut_bels = ['A6LUT', 'B6LUT', 'C6LUT', 'D6LUT']

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.csv', 'w')
f.write('module,loc,bel,n\n')
slices = gen_slices()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
for i in range(CLBN):
    bel = ''

    if random.randint(0, 1):
        module = 'clb_NCY0_MX'
    else:
        module = 'clb_NCY0_O5'
    n = random.randint(0, 3)
    loc = next(slices)
    bel = lut_bels[n]

    print('    %s' % module)
    print('            #(.LOC("%s"), .BEL("%s"), .N(%d))' % (loc, bel, n))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 8]), .dout(dout[  %d +: 8]));'
        % (i, 8 * i, 8 * i))

    f.write('%s,%s,%s,%s\n' % (module, loc, bel, n))
f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module clb_NCY0_MX (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_X16Y129_FIXME";
    parameter BEL="A6LUT_FIXME";
    parameter N=-1;

    wire [3:0] o;
    wire o6, o5;
    reg [3:0] s;

    always @(*) begin
        s = din[7:4];
        s[N] = o6;
    end

	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_0000_0000_0001)
	) lut (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(o5),
		.O6(o6));

	(* LOC=LOC, KEEP, DONT_TOUCH *)
    CARRY4 carry4(.O(o), .CO(), .DI(din[3:0]), .S(s), .CYINIT(1'b0), .CI());

	(* LOC=LOC, BEL=\"AFF\", KEEP, DONT_TOUCH *)
    FDRE fdce1(.D(o[0]), .C(clk), .CE(), .R(), .Q());
	(* LOC=LOC, BEL=\"BFF\", KEEP, DONT_TOUCH *)
    FDRE fdce2(.D(o[1]), .C(clk), .CE(), .R(), .Q());
	(* LOC=LOC, BEL=\"CFF\", KEEP, DONT_TOUCH *)
    FDRE fdce3(.D(o[2]), .C(clk), .CE(), .R(), .Q());
	(* LOC=LOC, BEL=\"DFF\", KEEP, DONT_TOUCH *)
    FDRE fdce4(.D(o[3]), .C(clk), .CE(), .R(), .Q());
endmodule

module clb_NCY0_O5 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_X16Y129_FIXME";
    parameter BEL="A6LUT_FIXME";
    parameter N=-1;

    wire [3:0] o;
    wire o6, o5;
    reg [3:0] s;
    reg [3:0] di;

    always @(*) begin
        s = din[7:4];
        s[N] = o6;

        di = {din[3:0]};
        di[N] = o5;
    end

	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_0000_0000_0001)
	) lut (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(o5),
		.O6(o6));

	(* LOC=LOC, KEEP, DONT_TOUCH *)
    CARRY4 carry4(.O(o), .CO(), .DI(di), .S(s), .CYINIT(1'b0), .CI());

	(* LOC=LOC, BEL=\"AFF\", KEEP, DONT_TOUCH *)
    FDRE fdce1(.D(o[0]), .C(clk), .CE(), .R(), .Q());
	(* LOC=LOC, BEL=\"BFF\", KEEP, DONT_TOUCH *)
    FDRE fdce2(.D(o[1]), .C(clk), .CE(), .R(), .Q());
	(* LOC=LOC, BEL=\"CFF\", KEEP, DONT_TOUCH *)
    FDRE fdce3(.D(o[2]), .C(clk), .CE(), .R(), .Q());
	(* LOC=LOC, BEL=\"DFF\", KEEP, DONT_TOUCH *)
    FDRE fdce4(.D(o[3]), .C(clk), .CE(), .R(), .Q());
endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import argparse
import json
from collections import OrderedDict

boolean_params = [
    ("AER_CAP_ECRC_CHECK_CAPABLE", 1),
    ("AER_CAP_ECRC_GEN_CAPABLE", 1),
    ("AER_CAP_PERMIT_ROOTERR_UPDATE", 1),
    ("AER_CAP_ON", 1),
    ("AER_CAP_MULTIHEADER", 1),
    ("CMD_INTX_IMPLEMENTED", 1),
    ("CPL_TIMEOUT_DISABLE_SUPPORTED", 1),
    ("DEV_CAP2_ARI_FORWARDING_SUPPORTED", 1),
    ("DEV_CAP2_ATOMICOP_ROUTING_SUPPORTED", 1),
    ("DEV_CAP2_ATOMICOP32_COMPLETER_SUPPORTED", 1),
    ("DEV_CAP2_ATOMICOP64_COMPLETER_SUPPORTED", 1),
    ("DEV_CAP2_CAS128_COMPLETER_SUPPORTED", 1),
    ("DEV_CAP2_NO_RO_ENABLED_PRPR_PASSING", 1),
    ("DEV_CAP2_LTR_MECHANISM_SUPPORTED", 1),
    ("DEV_CAP2_EXTENDED_FMT_FIELD_SUPPORTED", 1),
    ("DEV_CAP2_ENDEND_TLP_PREFIX_SUPPORTED", 1),
    ("ENDEND_TLP_PREFIX_FORWARDING_SUPPORTED", 1),
    ("DEV_CAP_ENABLE_SLOT_PWR_LIMIT_SCALE", 1),
    ("DEV_CAP_ENABLE_SLOT_PWR_LIMIT_VALUE", 1),
    ("DEV_CAP_EXT_TAG_SUPPORTED", 1),
    ("DEV_CAP_FUNCTION_LEVEL_RESET_CAPABLE", 1),
    ("DEV_CAP_ROLE_BASED_ERROR", 1),
    ("DEV_CONTROL_AUX_POWER_SUPPORTED", 1),
    ("DEV_CONTROL_EXT_TAG_DEFAULT", 1),
    ("DSN_CAP_ON", 1),
    ("INTERRUPT_STAT_AUTO", 1),
    ("IS_SWITCH", 1),
    ("LINK_CAP_CLOCK_POWER_MANAGEMENT", 1),
    ("LINK_CAP_DLL_LINK_ACTIVE_REPORTING_CAP", 1),
    ("LINK_CAP_LINK_BANDWIDTH_NOTIFICATION_CAP", 1),
    ("LINK_CAP_ASPM_OPTIONALITY", 1),
    ("LINK_CAP_SURPRISE_DOWN_ERROR_CAPABLE", 1),
    ("LINK_CTRL2_DEEMPHASIS", 1),
    ("LINK_CTRL2_HW_AUTONOMOUS_SPEED_DISABLE", 1),
    ("LINK_STATUS_SLOT_CLOCK_CONFIG", 1),
    ("MPS_FORCE", 1),
    ("MSI_CAP_64_BIT_ADDR_CAPABLE", 1),
    ("MSI_CAP_ON", 1),
    ("MSI_CAP_PER_VECTOR_MASKING_CAPABLE", 1),
    ("MSIX_CAP_ON", 1),
    ("PCIE_CAP_ON", 1),
    ("PCIE_CAP_SLOT_IMPLEMENTED", 1),
    ("PM_CAP_D1SUPPORT", 1),
    ("PM_CAP_D2SUPPORT", 1),
    ("PM_CAP_DSI", 1),
    ("PM_CAP_ON", 1),
    ("PM_CAP_PME_CLOCK", 1),
    ("PM_CSR_B2B3", 1),
    ("PM_CSR_BPCCEN", 1),
    ("PM_CSR_NOSOFTRST", 1),
    ("RBAR_CAP_ON", 1),
    ("ROOT_CAP_CRS_SW_VISIBILITY", 1),
    ("SELECT_DLL_IF", 1),
    ("SLOT_CAP_ATT_BUTTON_PRESENT", 1),
    ("SLOT_CAP_ATT_INDICATOR_PRESENT", 1),
    ("SLOT_CAP_ELEC_INTERLOCK_PRESENT", 1),
    ("SLOT_CAP_HOTPLUG_CAPABLE", 1),
    ("SLOT_CAP_HOTPLUG_SURPRISE", 1),
    ("SLOT_CAP_MRL_SENSOR_PRESENT", 1),
    ("SLOT_CAP_NO_CMD_COMPLETED_SUPPORT", 1),
    ("SLOT_CAP_POWER_CONTROLLER_PRESENT", 1),
    ("SLOT_CAP_POWER_INDICATOR_PRESENT", 1),
    ("SSL_MESSAGE_AUTO", 1),
    ("VC_CAP_ON", 1),
    ("VC_CAP_REJECT_SNOOP_TRANSACTIONS", 1),
    ("VSEC_CAP_IS_LINK_VISIBLE", 1),
    ("VSEC_CAP_ON", 1),
    ("LL_ACK_TIMEOUT_EN", 1),
    ("LL_REPLAY_TIMEOUT_EN", 1),
    ("PM_ASPML0S_TIMEOUT_EN", 1),
    ("PM_ASPM_FASTEXIT", 1),
    ("DISABLE_LANE_REVERSAL", 1),
    ("DISABLE_SCRAMBLING", 1),
    ("ENTER_RVRY_EI_L0", 1),
    ("ALLOW_X8_GEN2", 1),
    ("PL_FAST_TRAIN", 1),
    ("UPCONFIG_CAPABLE", 1),
    ("UPSTREAM_FACING", 1),
    ("EXIT_LOOPBACK_ON_EI", 1),
    ("DISABLE_ASPM_L1_TIMER", 1),
    ("DISABLE_BAR_FILTERING", 1),
    ("DISABLE_ID_CHECK", 1),
    ("DISABLE_RX_TC_FILTER", 1),
    ("DISABLE_RX_POISONED_RESP", 1),
    ("ENABLE_RX_TD_ECRC_TRIM", 1),
    ("TL_TFC_DISABLE", 1),
    ("TL_TX_CHECKS_DISABLE", 1),
    ("TL_RBYPASS", 1),
    ("DISABLE_PPM_FILTER", 1),
    ("DISABLE_LOCKED_FILTER", 1),
    ("USE_RID_PINS", 1),
    ("DISABLE_ERR_MSG", 1),
    ("PM_MF", 1),
    ("VC0_CPL_INFINITE", 1),
    ("RECRC_CHK_TRIM", 1),
    ("TECRC_EP_INV", 1),
    ("UR_INV_REQ", 1),
    ("UR_PRS_RESPONSE", 1),
    ("UR_ATOMIC", 1),
    ("UR_CFG1", 1),
    ("TRN_DW", 1),
    ("TRN_NP_FC", 1),
    ("USER_CLK2_DIV2", 1),
]

hex_params = [
    ("AER_CAP_ID", 16),
    ("AER_CAP_VERSION", 4),
    ("AER_BASE_PTR", 12),
    ("AER_CAP_NEXTPTR", 12),
    ("AER_CAP_OPTIONAL_ERR_SUPPORT", 24),
    ("BAR0", 32),
    ("BAR1", 32),
    ("BAR2", 32),
    ("BAR3", 32),
    ("BAR4", 32),
    ("BAR5", 32),
    ("EXPANSION_ROM", 32),
    ("CAPABILITIES_PTR", 8),
    ("CARDBUS_CIS_POINTER", 32),
    ("CLASS_CODE", 24),
    ("CPL_TIMEOUT_RANGES_SUPPORTED", 4),
    ("DEV_CAP2_TPH_COMPLETER_SUPPORTED", 2),
    ("DEV_CAP2_MAX_ENDEND_TLP_PREFIXES", 2),
    ("DSN_BASE_PTR", 12),
    ("DSN_CAP_ID", 16),
    ("DSN_CAP_NEXTPTR", 12),
    ("DSN_CAP_VERSION", 4),
    ("EXT_CFG_CAP_PTR", 6),
    ("EXT_CFG_XP_CAP_PTR", 10),
    ("HEADER_TYPE", 8),
    ("INTERRUPT_PIN", 8),
    ("LAST_CONFIG_DWORD", 10),
    ("LINK_CAP_MAX_LINK_SPEED", 4),
    ("LINK_CTRL2_TARGET_LINK_SPEED", 4),
    ("MSI_BASE_PTR", 8),
    ("MSI_CAP_ID", 8),
    ("MSI_CAP_NEXTPTR", 8),
    ("MSIX_BASE_PTR", 8),
    ("MSIX_CAP_ID", 8),
    ("MSIX_CAP_NEXTPTR", 8),
    ("MSIX_CAP_PBA_OFFSET", 29),
    ("MSIX_CAP_TABLE_OFFSET", 29),
    ("MSIX_CAP_TABLE_SIZE", 11),
    ("PCIE_BASE_PTR", 8),
    ("PCIE_CAP_CAPABILITY_ID", 8),
    ("PCIE_CAP_CAPABILITY_VERSION", 4),
    ("PCIE_CAP_DEVICE_PORT_TYPE", 4),
    ("PCIE_CAP_NEXTPTR", 8),
    ("PM_BASE_PTR", 8),
    ("PM_CAP_ID", 8),
    ("PM_CAP_NEXTPTR", 8),
    ("PM_CAP_PMESUPPORT", 5),
    ("PM_DATA_SCALE0", 2),
    ("PM_DATA_SCALE1", 2),
    ("PM_DATA_SCALE2", 2),
    ("PM_DATA_SCALE3", 2),
    ("PM_DATA_SCALE4", 2),
    ("PM_DATA_SCALE5", 2),
    ("PM_DATA_SCALE6", 2),
    ("PM_DATA_SCALE7", 2),
    ("PM_DATA0", 8),
    ("PM_DATA1", 8),
    ("PM_DATA2", 8),
    ("PM_DATA3", 8),
    ("PM_DATA4", 8),
    ("PM_DATA5", 8),
    ("PM_DATA6", 8),
    ("PM_DATA7", 8),
    ("RBAR_BASE_PTR", 12),
    ("RBAR_CAP_NEXTPTR", 12),
    ("RBAR_CAP_ID", 16),
    ("RBAR_CAP_VERSION", 4),
    ("RBAR_NUM", 3),
    ("RBAR_CAP_SUP0", 32),
    ("RBAR_CAP_SUP1", 32),
    ("RBAR_CAP_SUP2", 32),
    ("RBAR_CAP_SUP3", 32),
    ("RBAR_CAP_SUP4", 32),
    ("RBAR_CAP_SUP5", 32),
    ("RBAR_CAP_INDEX0", 3),
    ("RBAR_CAP_INDEX1", 3),
    ("RBAR_CAP_INDEX2", 3),
    ("RBAR_CAP_INDEX3", 3),
    ("RBAR_CAP_INDEX4", 3),
    ("RBAR_CAP_INDEX5", 3),
    ("RBAR_CAP_CONTROL_ENCODEDBAR0", 5),
    ("RBAR_CAP_CONTROL_ENCODEDBAR1", 5),
    ("RBAR_CAP_CONTROL_ENCODEDBAR2", 5),
    ("RBAR_CAP_CONTROL_ENCODEDBAR3", 5),
    ("RBAR_CAP_CONTROL_ENCODEDBAR4", 5),
    ("RBAR_CAP_CONTROL_ENCODEDBAR5", 5),
    ("SLOT_CAP_PHYSICAL_SLOT_NUM", 13),
    ("SLOT_CAP_SLOT_POWER_LIMIT_VALUE", 8),
    ("VC_BASE_PTR", 12),
    ("VC_CAP_NEXTPTR", 12),
    ("VC_CAP_ID", 16),
    ("VSEC_BASE_PTR", 12),
    ("VSEC_CAP_HDR_ID", 16),
    ("VSEC_CAP_HDR_LENGTH", 12),
    ("VSEC_CAP_HDR_REVISION", 4),
    ("VSEC_CAP_ID", 16),
    ("VSEC_CAP_NEXTPTR", 12),
    ("VSEC_CAP_VERSION", 4),
    ("CRM_MODULE_RSTS", 7),
    ("LL_ACK_TIMEOUT", 15),
    ("LL_REPLAY_TIMEOUT", 15),
    ("PM_ASPML0S_TIMEOUT", 15),
    ("INFER_EI", 5),
    ("LINK_CAP_MAX_LINK_WIDTH", 6),
    ("LTSSM_MAX_LINK_WIDTH", 6),
    ("DNSTREAM_LINK_NUM", 8),
    ("ENABLE_MSG_ROUTE", 11),
    ("VC_CAP_VERSION", 4),
    ("VC0_RX_RAM_LIMIT", 13),
    ("RP_AUTO_SPD", 2),
    ("RP_AUTO_SPD_LOOPCNT", 5),
    ("SPARE_BYTE0", 8),
    ("SPARE_BYTE1", 8),
    ("SPARE_BYTE2", 8),
    ("SPARE_BYTE3", 8),
    ("SPARE_WORD0", 32),
    ("SPARE_WORD1", 32),
    ("SPARE_WORD2", 32),
    ("SPARE_WORD3", 32),
]

int_params = [
    ("DEV_CAP_ENDPOINT_L0S_LATENCY", 3),
    ("DEV_CAP_ENDPOINT_L1_LATENCY", 3),
    ("DEV_CAP_MAX_PAYLOAD_SUPPORTED", 3),
    ("DEV_CAP_PHANTOM_FUNCTIONS_SUPPORT", 2),
    ("DEV_CAP_RSVD_14_12", 3),
    ("DEV_CAP_RSVD_17_16", 2),
    ("DEV_CAP_RSVD_31_29", 3),
    ("LINK_CAP_ASPM_SUPPORT", 2),
    ("LINK_CAP_L0S_EXIT_LATENCY_COMCLK_GEN1", 3),
    ("LINK_CAP_L0S_EXIT_LATENCY_COMCLK_GEN2", 3),
    ("LINK_CAP_L0S_EXIT_LATENCY_GEN1", 3),
    ("LINK_CAP_L0S_EXIT_LATENCY_GEN2", 3),
    ("LINK_CAP_L1_EXIT_LATENCY_COMCLK_GEN1", 3),
    ("LINK_CAP_L1_EXIT_LATENCY_COMCLK_GEN2", 3),
    ("LINK_CAP_L1_EXIT_LATENCY_GEN1", 3),
    ("LINK_CAP_L1_EXIT_LATENCY_GEN2", 3),
    ("LINK_CAP_RSVD_23", 1),
    ("LINK_CONTROL_RCB", 1),
    ("MSI_CAP_MULTIMSG_EXTENSION", 1),
    ("MSI_CAP_MULTIMSGCAP", 3),
    ("MSIX_CAP_PBA_BIR", 3),
    ("MSIX_CAP_TABLE_BIR", 3),
    ("PCIE_CAP_RSVD_15_14", 2),
    ("PCIE_REVISION", 4),
    ("PM_CAP_AUXCURRENT", 3),
    ("PM_CAP_RSVD_04", 1),
    ("PM_CAP_VERSION", 3),
    ("SLOT_CAP_SLOT_POWER_LIMIT_SCALE", 2),
    ("USER_CLK_FREQ", 3),
    ("LL_ACK_TIMEOUT_FUNC", 2),
    ("LL_REPLAY_TIMEOUT_FUNC", 2),
    ("PM_ASPML0S_TIMEOUT_FUNC", 2),
    ("N_FTS_COMCLK_GEN1", 8),
    ("N_FTS_COMCLK_GEN2", 8),
    ("N_FTS_GEN1", 8),
    ("N_FTS_GEN2", 8),
    ("PL_AUTO_CONFIG", 3),
    ("TL_RX_RAM_RADDR_LATENCY", 1),
    ("TL_RX_RAM_RDATA_LATENCY", 2),
    ("TL_RX_RAM_WRITE_LATENCY", 1),
    ("TL_TX_RAM_RADDR_LATENCY", 1),
    ("TL_TX_RAM_RDATA_LATENCY", 2),
    ("TL_TX_RAM_WRITE_LATENCY", 1),
    ("VC0_TOTAL_CREDITS_CD", 11),
    ("VC0_TOTAL_CREDITS_CH", 7),
    ("VC0_TOTAL_CREDITS_NPH", 7),
    ("VC0_TOTAL_CREDITS_NPD", 11),
    ("VC0_TOTAL_CREDITS_PD", 11),
    ("VC0_TOTAL_CREDITS_PH", 7),
    ("VC0_TX_LASTPACKET", 5),
    ("RECRC_CHK", 2),
    ("CFG_ECRC_ERR_CPLSTAT", 2),
    ("SPARE_BIT0", 1),
    ("SPARE_BIT1", 1),
    ("SPARE_BIT2", 1),
    ("SPARE_BIT3", 1),
    ("SPARE_BIT4", 1),
    ("SPARE_BIT5", 1),
    ("SPARE_BIT6", 1),
    ("SPARE_BIT7", 1),
    ("SPARE_BIT8", 1),
]


def dump_json():
    parser = argparse.ArgumentParser()
    parser.add_argument("--attrs-file", help="JSON output path", required=True)
    args = parser.parse_args()

    data = OrderedDict()

    # Prepare BOOL type attributes
    for param in boolean_params:
        data[param[0]] = {
            "type": "BOOL",
            "values": ["FALSE", "TRUE"],
            "digits": param[1]
        }

    # Prepare BIN type attributes
    for param in (hex_params + int_params):
        data[param[0]] = {
            "type": "BIN",
            "values": [(1 << param[1]) - 1],
            "digits": param[1]
        }

    data = dict(sorted(data.items()))

    # Emit JSON
    with open(args.attrs_file, "w") as f:
        json.dump(data, f, indent=4)


if __name__ == "__main__":
    dump_json()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Emits top.v's for various BUFHCE routing inputs. """
import os
import random
random.seed(int(os.getenv("SEED"), 16))
import re
from prjxray import util
from prjxray.lut_maker import LutMaker
from prjxray.db import Database
from io import StringIO

CMT_XY_FUN = util.create_xy_fun(prefix='')


def read_site_to_cmt():
    """ Yields clock sources and which CMT they route within. """
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt, tile = l.strip().split(',')
            yield (tile, site, cmt)


def make_ccio_route_options():

    # Read the PIP lists
    piplist_path = os.path.join(
        os.getenv("FUZDIR"), "..", "piplist", "build", "hclk_cmt")

    pips = []
    for fname in os.listdir(piplist_path):
        if not fname.endswith(".txt"):
            continue

        fullname = os.path.join(piplist_path, fname)
        with open(fullname, "r") as fp:
            pips += [l.strip() for l in fp.readlines()]

    # Get PIPs that mention FREQ_REFn wires. These are the ones that we want
    # force routing through.
    pips = [p for p in pips if "FREQ_REF" in p]

    # Sort by tile type
    options = {}
    for pip in pips:
        tile, dst, src = pip.split(".")

        for a, b in ((src, dst), (dst, src)):
            match = re.match(r".*FREQ_REF([0-3]).*", a)
            if match is not None:
                n = int(match.group(1))

                if tile not in options:
                    options[tile] = {}
                if n not in options[tile]:
                    options[tile][n] = set()

                options[tile][n].add(b)

    return options


class ClockSources(object):
    """ Class for tracking clock sources.

    Some clock sources can be routed to any CMT, for these, cmt='ANY'.
    For clock sources that belong to a CMT, cmt should be set to the CMT of
    the source.

    """

    def __init__(self):
        self.sources = {}
        self.source_to_cmt = {}
        self.used_sources_from_cmt = {}

    def add_clock_source(self, source, cmt):
        """ Adds a source from a specific CMT.

        cmt='ANY' indicates that this source can be routed to any CMT.
        """
        if cmt not in self.sources:
            self.sources[cmt] = []

        self.sources[cmt].append(source)
        self.source_to_cmt[source] = cmt

    def remove_clock_source(self, source, cmt="ANY"):
        """
        Removes a clock source from the available clock sources list
        """
        if source in self.source_to_cmt:
            del self.source_to_cmt[source]

        if cmt == "ANY":
            for sources in self.sources.values():
                if source in sources:
                    sources.remove(source)
        else:
            if source in self.sources[cmt]:
                self.sources[cmt].remove(source)

    def get_random_source(
            self, cmt, uses_left_right_routing=False, no_repeats=False):
        """ Get a random source that is routable to the specific CMT.

        get_random_source will return a source that is either cmt='ANY',
        cmt equal to the input CMT, or the adjecent CMT.

        """

        choices = []

        if cmt in self.sources:
            choices.extend(self.sources[cmt])

        if uses_left_right_routing:
            x, y = CMT_XY_FUN(cmt)

            if x % 2 == 0:
                x += 1
            else:
                x -= 1

            paired_cmt = 'X{}Y{}'.format(x, y)

            if paired_cmt in self.sources:
                for source in self.sources[paired_cmt]:
                    if 'BUFHCE' not in source:
                        choices.append(source)

        random.shuffle(choices)

        if not uses_left_right_routing:
            return choices[0]

        for source in choices:

            source_cmt = self.source_to_cmt[source]

            if source_cmt not in self.used_sources_from_cmt:
                self.used_sources_from_cmt[source_cmt] = set()

            if no_repeats and source in self.used_sources_from_cmt[source_cmt]:
                continue

            if len(self.used_sources_from_cmt[source_cmt]) >= 14:
                continue

            self.used_sources_from_cmt[source_cmt].add(source)

            return source

        return None


def get_paired_iobs(db, grid, tile_name):
    """ The two IOB33M's above and below the HCLK row have dedicate clock lines.
    """

    gridinfo = grid.gridinfo_at_tilename(tile_name)
    loc = grid.loc_of_tilename(tile_name)

    if gridinfo.tile_type.endswith('_L'):
        inc = 1
        lr = 'R'
    else:
        inc = -1
        lr = 'L'

    idx = 1
    while True:
        gridinfo = grid.gridinfo_at_loc((loc.grid_x + inc * idx, loc.grid_y))

        if gridinfo.tile_type.startswith('HCLK_IOI'):
            break

        idx += 1

    # A map of y deltas to CCIO wire indices
    CCIO_INDEX = {-1: 0, -3: 1, +2: 3, +4: 2}

    # Move from HCLK_IOI column to IOB column
    idx += 1

    for dy in [-1, -3, 2, 4]:
        iob_loc = (loc.grid_x + inc * idx, loc.grid_y + dy)
        gridinfo = grid.gridinfo_at_loc(iob_loc)
        tile_name = grid.tilename_at_loc(iob_loc)

        assert gridinfo.tile_type.startswith(lr + 'IOB'), (
            gridinfo, lr + 'IOB')

        for site, site_type in gridinfo.sites.items():
            if site_type in ['IOB33M', 'IOB18M']:
                yield tile_name, site, site_type[-3:-1], CCIO_INDEX[dy]


def check_allowed(mmcm_pll_dir, cmt):
    """ Check whether the CMT specified is in the allowed direction.

    This function is designed to bias sources to either the left or right
    input lines.

    """
    if mmcm_pll_dir == 'BOTH':
        return True
    elif mmcm_pll_dir == 'ODD':
        x, y = CMT_XY_FUN(cmt)
        return (x & 1) == 1
    elif mmcm_pll_dir == 'EVEN':
        x, y = CMT_XY_FUN(cmt)
        return (x & 1) == 0
    else:
        assert False, mmcm_pll_dir


def main():
    """
    HCLK_CMT switch box has the follow inputs:

    4 IOBs above and below
    14 MMCM outputs
    8 PLL outputs
    4 PHASER_IN outputs
    2 INT connections

    and the following outputs:

    3 PLLE2 inputs
    2 BUFMR inputs
    3 MMCM inputs
    ~2 MMCM -> BUFR???
    """

    clock_sources = ClockSources()
    adv_clock_sources = ClockSources()

    tile_site_cmt = list(read_site_to_cmt())
    site_to_cmt = {tsc[1]: tsc[2] for tsc in tile_site_cmt}
    cmt_to_hclk = {
        tsc[2]: tsc[0]
        for tsc in tile_site_cmt
        if tsc[0].startswith("HCLK_CMT_")
    }

    ccio_route_options = make_ccio_route_options()

    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    def gen_sites(desired_site_type):
        for tile_name in sorted(grid.tiles()):
            loc = grid.loc_of_tilename(tile_name)
            gridinfo = grid.gridinfo_at_loc(loc)
            for site, site_type in gridinfo.sites.items():
                if site_type == desired_site_type:
                    yield tile_name, site

    hclk_cmts = set()
    ins = []
    iobs = StringIO()

    hclk_cmt_tiles = set()
    for tile_name, site in gen_sites('BUFMRCE'):
        cmt = site_to_cmt[site]
        hclk_cmts.add(cmt)
        hclk_cmt_tiles.add(tile_name)

    mmcm_pll_only = random.randint(0, 1)
    mmcm_pll_dir = random.choice(('ODD', 'EVEN', 'BOTH'))

    print(
        '// mmcm_pll_only {} mmcm_pll_dir {}'.format(
            mmcm_pll_only, mmcm_pll_dir))

    have_iob_clocks = random.random() > .1

    iob_to_hclk = {}
    iob_clks = {}
    for tile_name in sorted(hclk_cmt_tiles):
        for _, site, volt, ccio in get_paired_iobs(db, grid, tile_name):
            iob_clock = 'clock_IBUF_{site}'.format(site=site)

            iob_to_hclk[site] = (tile_name, ccio)
            cmt = site_to_cmt[site]

            if cmt not in iob_clks:
                iob_clks[cmt] = ['']

            iob_clks[cmt].append(iob_clock)

            ins.append('input clk_{site}'.format(site=site))

            if have_iob_clocks:
                if check_allowed(mmcm_pll_dir, cmt):
                    clock_sources.add_clock_source(iob_clock, cmt)
                adv_clock_sources.add_clock_source(iob_clock, cmt)

            print(
                """
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    wire clock_IBUF_{site};
    IBUF #( .IOSTANDARD("LVCMOS{volt}") ) ibuf_{site} (
        .I(clk_{site}),
        .O(clock_IBUF_{site})
        );
                    """.format(volt=volt, site=site),
                file=iobs)

    print(
        '''
module top({inputs});
    (* KEEP, DONT_TOUCH *)
    LUT6 dummy();
    '''.format(inputs=', '.join(ins)))

    print(iobs.getvalue())

    luts = LutMaker()
    wires = StringIO()
    bufhs = StringIO()

    for _, site in gen_sites('MMCME2_ADV'):
        mmcm_clocks = [
            'mmcm_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(13)
        ]

        if check_allowed(mmcm_pll_dir, site_to_cmt[site]):
            for clk in mmcm_clocks:
                clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire cin1_{site}, cin2_{site}, clkfbin_{site}, {c0}, {c1}, {c2}, {c3}, {c4}, {c5};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    MMCME2_ADV pll_{site} (
        .CLKIN1(cin1_{site}),
        .CLKIN2(cin2_{site}),
        .CLKFBIN(clkfbin_{site}),
        .CLKOUT0({c0}),
        .CLKOUT0B({c1}),
        .CLKOUT1({c2}),
        .CLKOUT1B({c3}),
        .CLKOUT2({c4}),
        .CLKOUT2B({c5}),
        .CLKOUT3({c6}),
        .CLKOUT3B({c7}),
        .CLKOUT4({c8}),
        .CLKOUT5({c9}),
        .CLKOUT6({c10}),
        .CLKFBOUT({c11}),
        .CLKFBOUTB({c12})
    );
        """.format(
                site=site,
                c0=mmcm_clocks[0],
                c1=mmcm_clocks[1],
                c2=mmcm_clocks[2],
                c3=mmcm_clocks[3],
                c4=mmcm_clocks[4],
                c5=mmcm_clocks[5],
                c6=mmcm_clocks[6],
                c7=mmcm_clocks[7],
                c8=mmcm_clocks[8],
                c9=mmcm_clocks[9],
                c10=mmcm_clocks[10],
                c11=mmcm_clocks[11],
                c12=mmcm_clocks[12],
            ))

    for _, site in gen_sites('PLLE2_ADV'):
        pll_clocks = [
            'pll_clock_{site}_{idx}'.format(site=site, idx=idx)
            for idx in range(7)
        ]

        if check_allowed(mmcm_pll_dir, site_to_cmt[site]):
            for clk in pll_clocks:
                clock_sources.add_clock_source(clk, site_to_cmt[site])

        print(
            """
    wire cin1_{site}, cin2_{site}, clkfbin_{site}, {c0}, {c1}, {c2}, {c3}, {c4}, {c5}, {c6};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    PLLE2_ADV pll_{site} (
        .CLKIN1(cin1_{site}),
        .CLKIN2(cin2_{site}),
        .CLKFBIN(clkfbin_{site}),
        .CLKOUT0({c0}),
        .CLKOUT1({c1}),
        .CLKOUT2({c2}),
        .CLKOUT3({c3}),
        .CLKOUT4({c4}),
        .CLKOUT5({c5}),
        .CLKFBOUT({c6})
    );
        """.format(
                site=site,
                c0=pll_clocks[0],
                c1=pll_clocks[1],
                c2=pll_clocks[2],
                c3=pll_clocks[3],
                c4=pll_clocks[4],
                c5=pll_clocks[5],
                c6=pll_clocks[6],
            ))

    for tile_name, site in gen_sites('BUFHCE'):
        print(
            """
    wire I_{site};
    wire O_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFHCE buf_{site} (
        .I(I_{site}),
        .O(O_{site})
    );
                    """.format(site=site, ),
            file=bufhs)

        if site_to_cmt[site] in hclk_cmts:
            if not mmcm_pll_only:
                clock_sources.add_clock_source(
                    'O_{site}'.format(site=site), site_to_cmt[site])
            adv_clock_sources.add_clock_source(
                'O_{site}'.format(site=site), site_to_cmt[site])

    hclks_used_by_cmt = {}
    for cmt in site_to_cmt.values():
        hclks_used_by_cmt[cmt] = set()

    def check_hclk_src(src, src_cmt):
        if len(hclks_used_by_cmt[src_cmt]
               ) >= 12 and src not in hclks_used_by_cmt[src_cmt]:
            return None
        else:
            hclks_used_by_cmt[src_cmt].add(src)
            return src

    # Track used IOB sources
    used_iob_clks = set()

    if random.random() > .10:
        for tile_name, site in gen_sites('BUFHCE'):

            wire_name = clock_sources.get_random_source(
                site_to_cmt[site],
                uses_left_right_routing=True,
                no_repeats=mmcm_pll_only)

            if wire_name is not None and 'BUFHCE' in wire_name:
                # Looping a BUFHCE to a BUFHCE requires using a hclk in the
                # CMT of the source
                src_cmt = clock_sources.source_to_cmt[wire_name]

                wire_name = check_hclk_src(wire_name, src_cmt)

            if wire_name is None:
                continue

            if "IBUF" in wire_name:
                used_iob_clks.add(wire_name)
                clock_sources.remove_clock_source(wire_name)
                adv_clock_sources.remove_clock_source(wire_name)

            print(
                """
        assign I_{site} = {wire_name};""".format(
                    site=site,
                    wire_name=wire_name,
                ),
                file=bufhs)

    for tile_name, site in gen_sites('BUFMRCE'):
        pass

    for l in luts.create_wires_and_luts():
        print(l)

    print(wires.getvalue())
    print(bufhs.getvalue())

    for _, site in gen_sites('BUFR'):

        # Do not use BUFR always
        if random.random() < 0.50:
            continue

        available_srcs = set(iob_clks[site_to_cmt[site]]) - used_iob_clks
        if len(available_srcs) == 0:
            continue

        src = random.choice(list(available_srcs))

        if src != "":
            used_iob_clks.add(src)
            clock_sources.remove_clock_source(src)
            adv_clock_sources.remove_clock_source(src)

        adv_clock_sources.add_clock_source(
            'O_{site}'.format(site=site), site_to_cmt[site])

        print(
            """
    wire O_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFR bufr_{site} (
        .I({I}),
        .O(O_{site})
        );""".format(I=src, site=site))

    route_file = open("routes.txt", "w")

    def fix_ccio_route(net):

        # Get the IOB site name
        match = re.match(r".*_IBUF_(.*)", net)
        assert match is not None, net
        iob_site = match.group(1)

        # Get associated HCLK_CMT tile and CCIO wire index
        hclk_tile_name, ccio = iob_to_hclk[iob_site]

        # Get HCLK_CMT tile type
        hclk_tile = hclk_tile_name.rsplit("_", maxsplit=1)[0]

        # Pick a random route option
        opts = list(ccio_route_options[hclk_tile][ccio])
        route = random.choice(opts)
        route = "{}/{}".format(hclk_tile_name, route)
        route_file.write("{} {}\n".format(net, route))

    for _, site in gen_sites('PLLE2_ADV'):
        for cin in ('cin1', 'cin2', 'clkfbin'):
            if random.random() > .2:

                src = adv_clock_sources.get_random_source(site_to_cmt[site])

                src_cmt = adv_clock_sources.source_to_cmt[src]

                if 'IBUF' not in src and 'BUFR' not in src:
                    # Clocks from input pins do not require HCLK's, all other
                    # sources route from a global row clock.
                    src = check_hclk_src(src, src_cmt)

                if src is None:
                    continue

                if "IBUF" in src:
                    clock_sources.remove_clock_source(src)
                    adv_clock_sources.remove_clock_source(src)
                    fix_ccio_route(src)

                print(
                    """
        assign {cin}_{site} = {csrc};
            """.format(cin=cin, site=site, csrc=src))

    for _, site in gen_sites('MMCME2_ADV'):
        for cin in ('cin1', 'cin2', 'clkfbin'):
            if random.random() > .2:

                src = adv_clock_sources.get_random_source(site_to_cmt[site])

                src_cmt = adv_clock_sources.source_to_cmt[src]
                if 'IBUF' not in src and 'BUFR' not in src:
                    # Clocks from input pins do not require HCLK's, all other
                    # sources route from a global row clock.
                    src = check_hclk_src(src, src_cmt)

                if src is None:
                    continue

                if "IBUF" in src:
                    clock_sources.remove_clock_source(src)
                    adv_clock_sources.remove_clock_source(src)
                    fix_ccio_route(src)

                print(
                    """
        assign {cin}_{site} = {csrc};
            """.format(cin=cin, site=site, csrc=src))

    print("endmodule")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import xjson
import csv
import argparse
import sys
import fasm
from prjxray.db import Database
from prjxray.roi import Roi
from prjxray.util import get_db_root, get_part
from prjxray.xjson import extract_numbers


def set_port_wires(ports, name, pin, wires_outside_roi):
    for port in ports:
        if name == port['name']:
            port['wires_outside_roi'] = sorted(
                wires_outside_roi, key=extract_numbers)
            assert port['pin'] == pin
            return

    assert False, name


def main():
    parser = argparse.ArgumentParser(
        description=
        "Creates design.json from output of ROI generation tcl script.")
    parser.add_argument('--design_txt', required=True)
    parser.add_argument('--design_info_txt', required=True)
    parser.add_argument('--pad_wires', required=True)
    parser.add_argument('--design_fasm', required=True)

    args = parser.parse_args()

    design_json = {}
    design_json['ports'] = []
    design_json['info'] = {}
    with open(args.design_txt) as f:
        for d in csv.DictReader(f, delimiter=' '):
            if d['name'].startswith('dout['):
                d['type'] = 'out'
            elif d['name'].startswith('din['):
                d['type'] = 'in'
            elif d['name'].startswith('clk'):
                d['type'] = 'clk'
            else:
                assert False, d

            design_json['ports'].append(d)

    with open(args.design_info_txt) as f:
        for l in f:
            name, value = l.strip().split(' = ')

            design_json['info'][name] = int(value)

    db = Database(get_db_root(), get_part())
    grid = db.grid()

    roi = Roi(
        db=db,
        x1=design_json['info']['GRID_X_MIN'],
        y1=design_json['info']['GRID_Y_MIN'],
        x2=design_json['info']['GRID_X_MAX'],
        y2=design_json['info']['GRID_Y_MAX'],
    )

    with open(args.pad_wires) as f:
        for l in f:
            parts = l.strip().split(' ')
            name = parts[0]
            pin = parts[1]
            wires = parts[2:]

            wires_outside_roi = []

            for wire in wires:
                tile = wire.split('/')[0]

                loc = grid.loc_of_tilename(tile)

                if not roi.tile_in_roi(loc):
                    wires_outside_roi.append(wire)

            set_port_wires(design_json['ports'], name, pin, wires_outside_roi)

    frames_in_use = set()
    for tile in roi.gen_tiles():
        gridinfo = grid.gridinfo_at_tilename(tile)

        for bit in gridinfo.bits.values():
            frames_in_use.add(bit.base_address)

    required_features = []
    for fasm_line in fasm.parse_fasm_filename(args.design_fasm):
        if fasm_line.annotations:
            for annotation in fasm_line.annotations:
                if annotation.name != 'unknown_segment':
                    continue

                unknown_base_address = int(annotation.value, 0)

                assert False, "Found unknown bit in base address 0x{:08x}".format(
                    unknown_base_address)

        if not fasm_line.set_feature:
            continue

        tile = fasm_line.set_feature.feature.split('.')[0]

        loc = grid.loc_of_tilename(tile)
        gridinfo = grid.gridinfo_at_tilename(tile)

        not_in_roi = not roi.tile_in_roi(loc)

        if not_in_roi:
            required_features.append(fasm_line)

    design_json['required_features'] = sorted(
        fasm.fasm_tuple_to_string(required_features,
                                  canonical=True).split('\n'),
        key=extract_numbers)

    design_json['ports'].sort(key=lambda x: extract_numbers(x['name']))

    xjson.pprint(sys.stdout, design_json)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import argparse

from migen import *

from litex.boards.platforms import arty
from litex.soc.integration.soc_core import *
from litex.soc.integration.builder import *

from litex.soc.cores.clock import *

# CRG ----------------------------------------------------------------------------------------------


class _CRG(Module):
    def __init__(self, platform, sys_clk_freq):
        self.clock_domains.cd_sys = ClockDomain()

        # # #

        self.cd_sys.clk.attr.add("keep")

        self.submodules.pll = pll = S7PLL(speedgrade=-1)
        self.comb += pll.reset.eq(~platform.request("cpu_reset"))
        pll_clkin = Signal()
        pll.register_clkin(pll_clkin, 100e6)
        pll.create_clkout(self.cd_sys, sys_clk_freq)

        self.specials += Instance(
            "BUFG", i_I=platform.request("clk100"), o_O=pll_clkin)


# BaseSoC ------------------------------------------------------------------------------------------


class BaseSoC(SoCCore):
    def __init__(
            self, sys_clk_freq=int(50e6), integrated_rom_size=0x8000,
            **kwargs):
        platform = arty.Platform()
        SoCCore.__init__(
            self,
            platform,
            clk_freq=sys_clk_freq,
            integrated_rom_size=integrated_rom_size,
            integrated_sram_size=0x8000,
            ident="MiniLitex",
            cpu_variant="lite",
            **kwargs)

        self.submodules.crg = _CRG(platform, sys_clk_freq)


# Build --------------------------------------------------------------------------------------------


def main():
    parser = argparse.ArgumentParser(description="LiteX SoC on Arty")
    builder_args(parser)
    soc_core_args(parser)
    args = parser.parse_args()

    cls = BaseSoC
    soc = cls(**soc_core_argdict(args))
    builder = Builder(soc, **builder_argdict(args))
    builder.build()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

# Header
contrib = ["""\
# Contributing to Project X-Ray
"""]

# Extract the "Contributing" section from README.md
found = False
for line in open('README.md'):
    if found:
        if line.startswith('# '):
            # Found the next top level header
            break
        contrib.append(line)
    else:
        if line.startswith('# Contributing'):
            found = True

# Footer
contrib.append(
    """\






----

This file is generated from [README.md](README.md), please edit that file then
run the `./.github/update-contributing.py`.

""")

open("CONTRIBUTING.md", "w").write("".join(contrib))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import math
import os
import functools
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.verilog import vrandbits
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        if gridinfo.tile_type not in ['BRAM_L', 'BRAM_R']:
            continue

        sites = {}
        for site_name, site_type in gridinfo.sites.items():
            sites[site_type] = site_name

        yield tile_name, sites


@functools.lru_cache(maxsize=None)
def prepare_rand_int_choices(minval, maxval):
    """ Creates list ints between minval and maxval to allow fuzzer to uniquely identify all bits."""
    assert minval >= 0
    assert maxval >= minval

    min_p2 = math.floor(math.log(max(minval, 1), 2))
    max_p2 = math.ceil(math.log(maxval + 1, 2))

    if 2**max_p2 > maxval:
        max_search_p2 = max_p2 - 1
    else:
        max_search_p2 = max_p2

    choices = set(
        [minval, maxval, 2**(min_p2 + 1) - 1, 2**(max_search_p2) - 1])

    lb = min_p2
    ub = max_search_p2

    while lb < ub:
        ub = int(round(ub / 2.))

        val = 2**ub - 1

        lowval = val
        if lowval < minval:
            lowval |= (1 << max_search_p2)
        assert lowval >= minval, (val, ub)
        choices.add(lowval)

        highval = val << (max_search_p2 - ub)
        if highval > minval:
            assert highval <= maxval, (val, ub)
            choices.add(highval)

    for bit in range(max_search_p2):
        if 2**bit > minval:
            choices.add(2**bit)
        else:
            choices.add(2**bit | 2**max_search_p2)
            choices.add(2**bit | 2**(max_search_p2 - 1))

    zeros = set()
    ones = set()

    for choice in choices:
        assert choice >= minval, choice
        assert choice <= maxval, choice

        for bit in range(max_p2):
            if (1 << bit) & choice:
                ones.add(bit)
            else:
                zeros.add(bit)

    assert len(ones) == max_p2
    assert len(zeros) == max_p2

    return tuple(sorted(choices))


def rand_int(minval, maxval):
    return random.choice(prepare_rand_int_choices(minval, maxval))


def main():
    print('''
module top();
    ''')

    params_list = []
    for tile_name, sites in gen_sites():
        params = {}
        params['tile'] = tile_name
        params['site'] = sites['RAMBFIFO36E1']

        params['DATA_WIDTH'] = random.choice([4, 9, 18, 36, 72])
        params['EN_SYN'] = random.randint(0, 1)
        params['DO_REG'] = 1
        if params['EN_SYN']:
            params['FIRST_WORD_FALL_THROUGH'] = 0
        else:
            params['FIRST_WORD_FALL_THROUGH'] = random.randint(0, 1)

        if params['EN_SYN']:
            MIN_ALMOST_FULL_OFFSET = 1
            if params['DATA_WIDTH'] == 4:
                MAX_ALMOST_FULL_OFFSET = 8190
            elif params['DATA_WIDTH'] == 9:
                MAX_ALMOST_FULL_OFFSET = 4094
            elif params['DATA_WIDTH'] == 18:
                MAX_ALMOST_FULL_OFFSET = 2046
            elif params['DATA_WIDTH'] == 36:
                MAX_ALMOST_FULL_OFFSET = 1022
            elif params['DATA_WIDTH'] == 72:
                MAX_ALMOST_FULL_OFFSET = 510
            else:
                assert False

            MIN_ALMOST_EMPTY_OFFSET = MIN_ALMOST_FULL_OFFSET
            MAX_ALMOST_EMPTY_OFFSET = MAX_ALMOST_FULL_OFFSET
        else:
            MIN_ALMOST_FULL_OFFSET = 4
            if params['DATA_WIDTH'] == 4:
                MAX_ALMOST_FULL_OFFSET = 8185
            elif params['DATA_WIDTH'] == 9:
                MAX_ALMOST_FULL_OFFSET = 4089
            elif params['DATA_WIDTH'] == 18:
                MAX_ALMOST_FULL_OFFSET = 2041
            elif params['DATA_WIDTH'] == 36:
                MAX_ALMOST_FULL_OFFSET = 1017
            elif params['DATA_WIDTH'] == 72:
                MAX_ALMOST_FULL_OFFSET = 505
            else:
                assert False

            if params['FIRST_WORD_FALL_THROUGH']:
                MIN_ALMOST_EMPTY_OFFSET = MIN_ALMOST_FULL_OFFSET + 2
                MAX_ALMOST_EMPTY_OFFSET = MAX_ALMOST_FULL_OFFSET + 2
            else:
                MIN_ALMOST_EMPTY_OFFSET = MIN_ALMOST_FULL_OFFSET + 1
                MAX_ALMOST_EMPTY_OFFSET = MAX_ALMOST_FULL_OFFSET + 1

        ALMOST_EMPTY_OFFSET = rand_int(
            MIN_ALMOST_EMPTY_OFFSET, MAX_ALMOST_EMPTY_OFFSET)
        ALMOST_FULL_OFFSET = rand_int(
            MIN_ALMOST_FULL_OFFSET, MAX_ALMOST_FULL_OFFSET)
        params['ALMOST_EMPTY_OFFSET'] = "13'b{:013b}".format(
            ALMOST_EMPTY_OFFSET)
        params['ALMOST_FULL_OFFSET'] = "13'b{:013b}".format(ALMOST_FULL_OFFSET)

        if params['DATA_WIDTH'] == 36:
            params['FIFO_MODE'] = verilog.quote('FIFO36_72')
        else:
            params['FIFO_MODE'] = verilog.quote(
                'FIFO36_72'
            )  #verilog.quote('FIFO18') #verilog.quote(random.choice(('FIFO18', 'FIFO18_36')))

        params['INIT'] = '0'  #vrandbits(36)
        params['SRVAL'] = '0'  #vrandbits(36)

        print(
            '''
            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            FIFO36E1 #(
                .ALMOST_EMPTY_OFFSET({ALMOST_EMPTY_OFFSET}),
                .ALMOST_FULL_OFFSET({ALMOST_FULL_OFFSET}),
                .DATA_WIDTH({DATA_WIDTH}),
                .DO_REG({DO_REG}),
                .EN_SYN({EN_SYN}),
                .FIFO_MODE({FIFO_MODE}),
                .FIRST_WORD_FALL_THROUGH({FIRST_WORD_FALL_THROUGH}),
                .INIT({INIT}),
                .SRVAL({SRVAL})
                ) fifo_{site} (
                );
            '''.format(**params, ))

        params['FIFO_MODE'] = verilog.unquote(params['FIFO_MODE'])

        params_list.append(params)

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params_list, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker
from prjxray import verilog


def bitfilter(frame, word):
    if frame < 28:
        return False

    return True


def bus_tags(segmk, ps, site):
    segmk.add_site_tag(site, 'IN_USE', ps['active'])

    if not ps['active']:
        return

    for k in ps:
        segmk.add_site_tag(site, 'param_' + k + '_' + str(ps[k]), 1)

    for reg, invert in [
        ('RST', 1),
        ('PWRDWN', 1),
        ('CLKINSEL', 0),
    ]:
        opt = 'IS_{}_INVERTED'.format(reg)

        if invert:
            segmk.add_site_tag(site, 'ZINV_' + reg, 1 ^ ps[opt])
        else:
            segmk.add_site_tag(site, 'INV_' + reg, ps[opt])

    for opt in ['OPTIMIZED', 'HIGH', 'LOW']:
        if verilog.unquote(ps['BANDWIDTH']) == opt:
            segmk.add_site_tag(site, 'BANDWIDTH.' + opt, 1)
        elif verilog.unquote(ps['BANDWIDTH']) == 'LOW':
            segmk.add_site_tag(site, 'BANDWIDTH.' + opt, 0)

    for opt in ['ZHOLD', 'BUF_IN', 'EXTERNAL', 'INTERNAL']:
        continue

        opt_match = verilog.unquote(ps['COMPENSATION']) == opt

        if ps['clkfbin_conn'] == '':
            segmk.add_site_tag(site, 'COMP.NOFB_' + opt, opt_match)
            segmk.add_site_tag(site, 'COMP.ZNOFB_' + opt, opt_match)
            continue

        for conn in ['clk', 'clkfbout_mult_BUFG_' + ps['site'],
                     'clkfbout_mult_' + ps['site']]:
            conn_match = ps['clkfbin_conn'] == conn
            segmk.add_site_tag(
                site, 'COMP.' + opt + '_' + conn + '_' + ps['site'], opt_match
                and conn_match)
            segmk.add_site_tag(
                site, 'COMP.Z' + opt + '_' + conn + '_' + ps['site'],
                not opt_match and conn_match)
            segmk.add_site_tag(
                site, 'COMP.Z' + opt + '_Z' + conn + '_' + ps['site'],
                not opt_match and not conn_match)
            segmk.add_site_tag(
                site, 'COMP.' + opt + '_Z' + conn + '_' + ps['site'], opt_match
                and not conn_match)

    bufg_on_clkin = \
            'BUFG' in ps['clkin1_conn'] or \
            'BUFG' in ps['clkin2_conn']

    # This one is in conflict with some clock routing bits.
    #    match = verilog.unquote(ps['COMPENSATION']) in ['BUF_IN', 'EXTERNAL']
    #    if not match:
    #        if verilog.unquote(ps['COMPENSATION']) == 'ZHOLD' and bufg_on_clkin:
    #            match = True
    #    segmk.add_site_tag(
    #        site, 'COMPENSATION.BUF_IN_OR_EXTERNAL_OR_ZHOLD_CLKIN_BUF', match)

    match = verilog.unquote(ps['COMPENSATION']) in ['ZHOLD']
    segmk.add_site_tag(
        site, 'COMPENSATION.Z_ZHOLD_OR_CLKIN_BUF', not match
        or (match and bufg_on_clkin))
    segmk.add_site_tag(
            site, 'COMPENSATION.ZHOLD_NO_CLKIN_BUF', match and \
                    not bufg_on_clkin
                    )
    segmk.add_site_tag(
            site, 'COMPENSATION.ZHOLD_NO_CLKIN_BUF_NO_TOP', match and \
                    not bufg_on_clkin and \
                    site != "PLLE2_ADV_X0Y3" and site != "PLLE2_ADV_X0Y0"
                    )
    segmk.add_site_tag(
            site, 'COMP.ZHOLD_NO_CLKIN_BUF_TOP', match and \
                    not bufg_on_clkin and \
                    (site == "PLLE2_ADV_X0Y3" or site == "PLLE2_ADV_X0Y0")
                    )

    # No INTERNAL as it has conflicting bits
    for opt in ['ZHOLD', 'BUF_IN', 'EXTERNAL']:
        if opt in ['BUF_IN', 'EXTERNAL']:
            if ps['clkfbin_conn'] not in ['', 'clk']:
                continue

        if site == "PLLE2_ADV_X0Y2" and opt == 'ZHOLD':
            segmk.add_site_tag(
                site, 'TOP.COMPENSATION.' + opt,
                verilog.unquote(ps['COMPENSATION']) == opt)
        else:
            segmk.add_site_tag(
                site, 'COMPENSATION.' + opt,
                verilog.unquote(ps['COMPENSATION']) == opt)
        segmk.add_site_tag(
            site, 'COMPENSATION.Z_' + opt,
            verilog.unquote(ps['COMPENSATION']) != opt)


# This one has bits that are in conflict with clock routing
#    segmk.add_site_tag(
#        site, 'COMPENSATION.INTERNAL',
#        verilog.unquote(ps['COMPENSATION']) in ['INTERNAL'])

    for param in ['CLKFBOUT_MULT']:
        paramadj = int(ps[param])
        bitstr = [int(x) for x in "{0:09b}".format(paramadj)[::-1]]
        for i in range(7):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])

    for param in ['CLKOUT0_DUTY_CYCLE']:
        assert ps[param][:2] == '0.', ps[param]
        assert len(ps[param]) == 5
        paramadj = int(ps[param][2:])
        bitstr = [int(x) for x in "{0:011b}".format(paramadj)[::-1]]

        for i in range(10):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])

    for param, bits in [
        ('CLKOUT0_DIVIDE', 7),
        ('CLKOUT1_DIVIDE', 7),
        ('CLKOUT2_DIVIDE', 7),
        ('CLKOUT3_DIVIDE', 7),
        ('CLKOUT4_DIVIDE', 7),
        ('CLKOUT5_DIVIDE', 7),
        ('DIVCLK_DIVIDE', 6),
    ]:
        # 1-128 => 0-127 for actual 7 bit value
        paramadj = int(ps[param])
        if paramadj < 4:
            continue

        bitstr = [int(x) for x in "{0:09b}".format(paramadj)[::-1]]
        for i in range(bits):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])

    segmk.add_site_tag(
        site, 'STARTUP_WAIT',
        verilog.unquote(ps['STARTUP_WAIT']) == 'TRUE')


def run():

    segmk = Segmaker("design.bits")

    print("Loading tags")
    f = open('params.jl', 'r')
    f.readline()
    for l in f:
        j = json.loads(l)
        bus_tags(segmk, j, j['site'])

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
Python 3 removed the 'cmp' function and raises a Type error when you try to
compare different types. This module recreates Python 2 style 'cmp' function
which produces a "total ordering" for mixed type lists.
"""

import functools
import itertools


def cmp(a, b):
    """

    >>> cmp(1, 1)
    0
    >>> cmp('A', 'A')
    0
    >>> cmp(None, None)
    0
    >>> cmp(('A', 'B'), ('A', 'B'))
    0
    >>> cmp(['A', 'B'], ('A', 'B'))
    0
    >>> cmp((1, 2), (1, 2))
    0
    >>> cmp((1, 2), [1, 2])
    0

    >>> cmp(1, 2)
    -1
    >>> cmp('A', 'B')
    -1
    >>> cmp(('A', 'B'), ('A', 'C'))
    -1
    >>> cmp(['A', 'B'], ('A', 'C'))
    -1
    >>> cmp((1, 2), (1, 3))
    -1
    >>> cmp((1, 2), [1, 3])
    -1

    >>> cmp(2, 1)
    1
    >>> cmp('B', 'A')
    1
    >>> cmp(('A', 'C'), ('A', 'B'))
    1
    >>> cmp(['A', 'C'], ('A', 'B'))
    1
    >>> cmp((1, 3), (1, 2))
    1
    >>> cmp((1, 3), [1, 2])
    1

    >>> cmp(1, None)
    1
    >>> cmp('A', None)
    1
    >>> cmp(('A', 'B'), None)
    1
    >>> cmp(['A', 'B'], None)
    1
    >>> cmp((1, 2), None)
    1
    >>> cmp((1, 2), None)
    1

    >>> cmp(None, 2)
    -1
    >>> cmp(None, 'B')
    -1
    >>> cmp(None, ('A', 'B'))
    -1
    >>> cmp(None, ('A', 'C'))
    -1
    >>> cmp(None, (1, 2))
    -1
    >>> cmp(None, [1, 3])
    -1

    >>> cmp(1, 'A')
    -1
    >>> cmp('A', 1)
    1

    >>> cmp(('A', 'B'), 1)
    1
    >>> cmp(1, ['A', 'B'])
    -1

    >>> cmp((1, 2), 1)
    1
    >>> cmp(1, (1, 2))
    -1

    >>> cmp('A', 'AA')
    -1
    >>> cmp('AA', 'A')
    1

    >>> cmp(b'A', b'A')
    0
    >>> cmp(b'A', b'AA')
    -1
    >>> cmp(b'AA', b'A')
    1

    >>> def bit(*args):
    ...   return args
    >>> a = ('CLBLL', 'L', 'SLICEL', ('X', 0), 'AFFMUX', 'XOR')
    >>> b = ('CLBLL', 'L', 'SLICEL', ('X', 0), 'AFFMUX', ('F', 7))
    >>> cmp(a, b)
    -1
    >>> cmp(b, a)
    1

    """
    if not isinstance(a, (str, bytes)) and not isinstance(b, (str, bytes)):
        try:
            for i, j in itertools.zip_longest(iter(a), iter(b)):
                r = cmp(i, j)
                if r != 0:
                    return r
            return 0
        except TypeError:
            pass
    if type(a) == type(b):
        if a == b:
            return 0
        elif a < b:
            return -1
        elif a > b:
            return 1
        else:
            raise SystemError
    return cmp(a.__class__.__name__, b.__class__.__name__)


cmp_key = functools.cmp_to_key(cmp)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
#
# Produces a complete database of wires in the ROI and their connections and tests if each
# routing node is a tree (i.e. fails with an error when a loop is found).

import os, sys, json

from prjxray.util import OpenSafeFile

def main():
    with OpenSafeFile("%s/%s/tilegrid.json" % (os.getenv("XRAY_DATABASE_DIR"),
                                       os.getenv("XRAY_DATABASE")), "r") as f:
        tilegrid = json.load(f)

    with OpenSafeFile("%s/%s/tileconn.json" % (os.getenv("XRAY_DATABASE_DIR"),
                                       os.getenv("XRAY_DATABASE")), "r") as f:
        tileconn = json.load(f)

    grid = dict()
    wirepairs = dict()
    remwires = set()

    for tilename, tiledata in tilegrid.items():
        key = (tiledata["grid_x"], tiledata["grid_y"])
        grid[key] = tilename

    def check_tileconn_entry(tilename, tiledata, entry, idx):
        if idx == 0:
            otheridx = 1
            otherxy = (
                tiledata["grid_x"] + entry["grid_deltas"][0],
                tiledata["grid_y"] + entry["grid_deltas"][1])
        else:
            otheridx = 0
            otherxy = (
                tiledata["grid_x"] - entry["grid_deltas"][0],
                tiledata["grid_y"] - entry["grid_deltas"][1])

        if otherxy not in grid:
            return

        othertilename = grid[otherxy]
        othertiledata = tilegrid[othertilename]

        if othertiledata["type"] != entry["tile_types"][otheridx]:
            return

        print(
            "  Found relevant entry (%s %s %d %d): %s" % (
                entry["tile_types"][0], entry["tile_types"][1],
                entry["grid_deltas"][0], entry["grid_deltas"][1],
                othertilename))

        for pair in entry["wire_pairs"]:
            wirename = "%s/%s" % (tilename, pair[idx])
            otherwirename = "%s/%s" % (othertilename, pair[otheridx])

            remwires.add(wirename)
            remwires.add(otherwirename)

            if wirename not in wirepairs:
                wirepairs[wirename] = set()
            wirepairs[wirename].add(otherwirename)

            if otherwirename not in wirepairs:
                wirepairs[otherwirename] = set()
            wirepairs[otherwirename].add(wirename)

    for tilename, tiledata in tilegrid.items():
        print("Connecting wires in tile %s." % tilename)
        for entry in tileconn:
            if entry["tile_types"][0] == tiledata["type"]:
                check_tileconn_entry(tilename, tiledata, entry, 0)
            if entry["tile_types"][1] == tiledata["type"]:
                check_tileconn_entry(tilename, tiledata, entry, 1)

    def check_wire(wire, source):
        for next_wire in wirepairs[wire]:
            if next_wire == source:
                continue
            print("  %s" % next_wire)
            if next_wire not in remwires:
                print("  ** FOUND LOOP IN THIS NODE **")
                sys.exit(1)
            remwires.remove(next_wire)
            check_wire(next_wire, wire)

    while len(remwires):
        wire = remwires.pop()
        print("Checking %s:" % wire)
        check_wire(wire, None)

    print("NO LOOP FOUND: OK")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray.db import Database
from prjxray import util
from prjxray.state_gen import StateGen


def gen_sites():
    xy_fun = util.create_xy_fun('BUFGCTRL_')
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        sites = []

        xs = []
        ys = []
        for site, site_type in gridinfo.sites.items():
            if site_type == 'BUFGCTRL':
                x, y = xy_fun(site)
                xs.append(x)
                ys.append(y)

                sites.append((site, x, y))

        if sites:
            yield tile_name, min(xs), min(ys), sorted(sites)


def main():
    print('''
module top();
    ''')

    params_list = []
    state_gen = StateGen(list(gen_sites()), states_per_site=12 * 16)
    for tile_name, x_min, y_min, sites in state_gen:
        for site, x, y in sites:
            params = {}
            params['tile'] = tile_name
            params['site'] = site
            params['x'] = x - x_min
            params['y'] = y - y_min
            params['IN_USE'] = state_gen.next_state()

            if params['IN_USE']:
                params['INIT_OUT'] = state_gen.next_state()
                params['IS_CE0_INVERTED'] = state_gen.next_state()
                params['IS_CE1_INVERTED'] = state_gen.next_state()
                params['IS_S0_INVERTED'] = state_gen.next_state()
                params['IS_S1_INVERTED'] = state_gen.next_state()
                params['IS_IGNORE0_INVERTED'] = state_gen.next_state()
                params['IS_IGNORE1_INVERTED'] = state_gen.next_state()

                params['PRESELECT_I0'] = state_gen.next_state()
                if not params['PRESELECT_I0']:
                    params['PRESELECT_I1'] = state_gen.next_state()
                else:
                    params['PRESELECT_I1'] = 0

                params['force_0_high'] = state_gen.next_state()
                params['force_1_high'] = state_gen.next_state()
                if params['force_0_high']:
                    params['force_1_high'] = 0

                print(
                    '''
    wire ce0_{site};
    wire s0_{site};
    (* KEEP, DONT_TOUCH *)
    LUT6 l0_{site} (
        .O(ce0_{site})
        );
    (* KEEP, DONT_TOUCH *)
    LUT6 l1_{site} (
        .O(s0_{site})
        );

    wire ce1_{site};
    wire s1_{site};
    (* KEEP, DONT_TOUCH *)
    LUT6 l2_{site} (
        .O(ce1_{site})
        );
    (* KEEP, DONT_TOUCH *)
    LUT6 l3_{site} (
        .O(s1_{site})
        );

    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFGCTRL #(
        .INIT_OUT({INIT_OUT}),
        .PRESELECT_I0({PRESELECT_I0}),
        .PRESELECT_I1({PRESELECT_I1}),
        .IS_CE0_INVERTED({IS_CE0_INVERTED}),
        .IS_CE1_INVERTED({IS_CE1_INVERTED}),
        .IS_S0_INVERTED({IS_S0_INVERTED}),
        .IS_S1_INVERTED({IS_S1_INVERTED}),
        .IS_IGNORE0_INVERTED({IS_IGNORE0_INVERTED}),
        .IS_IGNORE1_INVERTED({IS_IGNORE1_INVERTED})
        ) buf_{site} (
            .CE0(ce0_{site} | {force_0_high}),
            .S0(s0_{site} | {force_0_high}),
            .CE1(ce1_{site} | {force_1_high}),
            .S1(s1_{site} | {force_1_high})
        );
                    '''.format(**params))

            params_list.append(params)

    print("""
    (* KEEP, DONT_TOUCH *)
    LUT6 dummy (
        );""")

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params_list, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import argparse
import json
from sdf_timing import sdfparse
from sdf_timing import utils


def generate_sdf(timings, sdffile):

    sdf_data = sdfparse.emit(timings, timescale='1ns')
    with open(sdffile, 'w') as fp:
        fp.write(sdf_data)


def add_timing_paths_entry(paths, type, values):
    paths[type] = dict()
    paths[type]['min'] = values[0]
    paths[type]['avg'] = values[1]
    paths[type]['max'] = values[2]
    return paths


def read_raw_timings(fin, site):

    timings = dict()
    timings['cells'] = dict()
    with open(fin, "r") as f:
        for line in f:

            raw_data = line.split()
            speed_model = raw_data[0]

            if speed_model.startswith('bel_d_'):
                speed_model = speed_model[6:]

            speed_model_split = speed_model.split('_')
            interconn_input = "_".join(speed_model_split[1:-1])
            interconn_output = speed_model_split[-1]
            celltype = "ROUTING_BEL"

            if celltype not in timings['cells']:
                timings['cells'][celltype] = dict()

            cellsite = site + '/' + interconn_output.upper()

            if cellsite not in timings['cells'][celltype]:
                timings['cells'][celltype][cellsite] = dict()

            if speed_model not in timings['cells'][celltype][cellsite]:
                timings['cells'][celltype][cellsite][speed_model] = dict()

            delays = dict()
            # each timing entry reports 5 delays
            for d in range(0, 5):
                (t, v) = raw_data[d + 1].split(':')
                delays[t] = v

            # create entry for sdf writer
            iport = dict()
            iport['port'] = interconn_input
            iport['port_edge'] = None
            oport = dict()
            oport['port'] = interconn_output
            oport['port_edge'] = None
            paths = dict()
            paths = add_timing_paths_entry(
                paths, 'slow', [delays['SLOW_MIN'], None, delays['SLOW_MAX']])
            paths = add_timing_paths_entry(
                paths, 'fast', [delays['FAST_MIN'], None, delays['FAST_MAX']])

            if speed_model.endswith('diff'):
                iport['port'] = "_".join(speed_model_split[1:])
                iport['port_edge'] = None
                timings['cells'][celltype][cellsite][
                    speed_model] = utils.add_device(iport, paths)
            else:
                timings['cells'][celltype][cellsite][
                    speed_model] = utils.add_interconnect(iport, oport, paths)
            timings['cells'][celltype][cellsite][speed_model][
                'is_absolute'] = True

    return timings


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--timings', type=str, help='Raw timing input file')
    parser.add_argument('--sdf', type=str, help='output sdf file')
    parser.add_argument(
        '--site', type=str, help='Site of the processed timings')
    parser.add_argument(
        '--debug', type=bool, default=False, help='Enable debug json dumps')
    args = parser.parse_args()

    timings = read_raw_timings(args.timings, args.site)
    if args.debug:
        with open("debug" + args.site + ".json", 'w') as fp:
            json.dump(timings, fp, indent=4, sort_keys=True)

    generate_sdf(timings, args.sdf)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
class Roi(object):
    """ Object that represents a Project X-ray ROI.

    Can be used to iterate over tiles and sites within an ROI.

    """

    def __init__(self, db, x1, x2, y1, y2):
        self.grid = db.grid()
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2

    def tile_in_roi(self, grid_loc):
        """ Returns true if grid_loc (GridLoc tuple) is within the ROI. """
        x = grid_loc.grid_x
        y = grid_loc.grid_y
        return self.x1 <= x and x <= self.x2 and self.y1 <= y and y <= self.y2

    def gen_tiles(self, tile_types=None):
        ''' Yield tile names within ROI.

        tile_types: list of tile types to keep, or None for all
        '''

        for tile_name in self.grid.tiles():
            loc = self.grid.loc_of_tilename(tile_name)

            if not self.tile_in_roi(loc):
                continue

            gridinfo = self.grid.gridinfo_at_loc(loc)

            if tile_types is not None and gridinfo.tile_type not in tile_types:
                continue

            yield tile_name

    def gen_sites(self, site_types=None):
        ''' Yield (tile_name, site_name, site_type) within ROI.

        site_types: list of site types to keep, or None for all

        '''

        for tile_name in self.grid.tiles():
            loc = self.grid.loc_of_tilename(tile_name)

            if not self.tile_in_roi(loc):
                continue

            gridinfo = self.grid.gridinfo_at_loc(loc)

            for site_name, site_type in gridinfo.sites.items():
                if site_types is not None and site_type not in site_types:
                    continue

                yield (tile_name, site_name, site_type)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
def ones(l):
    #return l + [x + '_1' for x in l]
    #return sorted(l + [x + '_1' for x in l])
    ret = []
    for x in l:
        ret.append(x)
        ret.append(x + '_1')
    return ret


# The complete primitive sets
ffprims_fall = ones(
    [
        'FD',
        'FDC',
        'FDCE',
        'FDE',
        'FDP',
        'FDPE',
        'FDR',
        'FDRE',
        'FDS',
        'FDSE',
    ])
ffprims_lall = ones([
    'LDC',
    'LDCE',
    'LDE',
    'LDPE',
    'LDP',
])

# Base primitives
ffprims_f = [
    'FDRE',
    'FDSE',
    'FDCE',
    'FDPE',
]
ffprims_l = [
    'LDCE',
    'LDPE',
]
ffprims = ffprims_f + ffprims_l


def isff(prim):
    return prim.startswith("FD")


def isl(prim):
    return prim.startswith("LD")


ff_bels_5 = [
    'A5FF',
    'B5FF',
    'C5FF',
    'D5FF',
]
ff_bels_ffl = [
    'AFF',
    'BFF',
    'CFF',
    'DFF',
]
ff_bels = ff_bels_ffl + ff_bels_5
#ff_bels = ff_bels_ffl



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker
from prjxray import verilog


def bus_tags(segmk, ps, site):
    for param in ['INIT_43']:
        paramadj = int(ps[param])
        bitstr = [int(x) for x in "{0:016b}".format(paramadj)[::-1]]
        for i in range(len(bitstr)):
            segmk.add_site_tag(site, '%s[%u]' % (param, i), bitstr[i])


def run():

    segmk = Segmaker("design.bits")

    print("Loading tags")
    f = open('params.jl', 'r')
    f.readline()
    for l in f:
        j = json.loads(l)
        ps = j['params']
        assert j['module'] == 'my_XADC'
        site = verilog.unquote(ps['LOC'])

        bus_tags(segmk, ps, site)

    segmk.compile()
    segmk.write()


run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Take bitstream .bit files and decode them to FASM.
'''

import contextlib
import os
import fasm
import fasm.output
from prjxray import fasm_disassembler
from prjxray import bitstream
from prjxray.db import Database
from prjxray.util import OpenSafeFile
import subprocess
import tempfile


def bit_to_bits(bitread, part_yaml, bit_file, bits_file, frame_range=None):
    """ Calls bitread to create bits (ASCII) from bit file (binary) """
    if frame_range:
        frame_range_arg = '-F {}'.format(frame_range)
    else:
        frame_range_arg = ''

    subprocess.check_output(
        '{} --part_file {} {} -o {} -z -y {}'.format(
            bitread, part_yaml, frame_range_arg, bits_file, bit_file),
        shell=True)


def bits_to_fasm(db_root, part, bits_file, verbose, canonical):
    db = Database(db_root, part)
    grid = db.grid()
    disassembler = fasm_disassembler.FasmDisassembler(db)

    with OpenSafeFile(bits_file) as f:
        bitdata = bitstream.load_bitdata(f)

    model = fasm.output.merge_and_sort(
        disassembler.find_features_in_bitstream(bitdata, verbose=verbose),
        zero_function=disassembler.is_zero_feature,
        sort_key=grid.tile_key,
    )

    print(fasm.fasm_tuple_to_string(model, canonical=canonical), end='')


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description='Convert 7-series bit file to FASM.')

    database_dir = os.getenv("XRAY_DATABASE_DIR")
    database = os.getenv("XRAY_DATABASE")
    db_root_kwargs = {}
    if database_dir is None or database is None:
        db_root_kwargs['required'] = True
    else:
        db_root_kwargs['required'] = False
        db_root_kwargs['default'] = os.path.join(database_dir, database)

    default_part = os.getenv("XRAY_PART")
    part_kwargs = {}
    if default_part is None:
        part_kwargs['required'] = True
    else:
        part_kwargs['required'] = False
        part_kwargs['default'] = default_part

    if os.getenv("XRAY_TOOLS_DIR") is None:
        default_bitread = 'bitread'
    else:
        default_bitread = os.path.join(os.getenv("XRAY_TOOLS_DIR"), 'bitread')

    parser.add_argument('--db-root', help="Database root.", **db_root_kwargs)
    parser.add_argument(
        '--bits-file',
        help="Output filename for bitread output, default is deleted tempfile.",
        default=None)
    parser.add_argument(
        '--part', help="Name of part being targetted.", **part_kwargs)
    parser.add_argument(
        '--bitread',
        help="bitread tool to to use",
        default=default_bitread)
    parser.add_argument(
        '--frame_range', help="Frame range to use with bitread.")
    parser.add_argument('bit_file', help='')
    parser.add_argument(
        '--verbose',
        help='Print lines for unknown tiles and bits',
        action='store_true')
    parser.add_argument(
        '--canonical', help='Output canonical bitstream.', action='store_true')
    args = parser.parse_args()

    with contextlib.ExitStack() as stack:
        if args.bits_file:
            bits_file = stack.enter_context(open(args.bits_file, 'wb'))
        else:
            bits_file = stack.enter_context(tempfile.NamedTemporaryFile())

        bit_to_bits(
            bitread=args.bitread,
            part_yaml=os.path.join(args.db_root, args.part, "part.yaml"),
            bit_file=args.bit_file,
            bits_file=bits_file.name,
            frame_range=args.frame_range,
        )

        bits_to_fasm(
            args.db_root, args.part, bits_file.name, args.verbose,
            args.canonical)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
import os
import os.path


def bitfilter(frame, word):
    if frame < 25 or frame > 29:
        return False

    return True


def merge_lr_wires(wire):
    wire = wire.replace('CMT_L_LOWER_B', 'CMT_LRMAP_LOWER_B')
    wire = wire.replace('CMT_R_LOWER_B', 'CMT_LRMAP_LOWER_B')
    return wire


def main():
    segmk = Segmaker("design.bits")

    designdata = {}
    tiledata = {}
    pipdata = {}
    ppipdata = {}
    ignpip = set()
    all_clks = {}

    piplists = ['cmt_top_l_lower_b.txt', 'cmt_top_r_lower_b.txt']
    wirelists = ['cmt_top_l_lower_b_wires.txt', 'cmt_top_r_lower_b_wires.txt']
    ppiplists = ['ppips_cmt_top_l_lower_b.db', 'ppips_cmt_top_r_lower_b.db']

    # Load PIP lists
    print("Loading PIP lists...")
    for piplist in piplists:
        with open(os.path.join(os.getenv('FUZDIR'), '..', 'piplist', 'build',
                               'cmt_top_lower', piplist)) as f:
            for l in f:
                pip, is_directional = l.strip().split(' ')
                tile_type, dst, src = pip.split('.')
                if tile_type not in pipdata:
                    pipdata[tile_type] = []
                    all_clks[tile_type] = set()

                pipdata[tile_type].append((src, dst))
                if dst.split('_')[-1].startswith('CLK'):
                    all_clks[tile_type].add(src)

                if not int(is_directional):
                    pipdata[tile_type].append((dst, src))
                    if src.split('_')[-1].startswith('CLK'):
                        all_clks[tile_type].add(dst)

    wiredata = {}
    for wirelist in wirelists:
        with open(os.path.join(os.getenv('FUZDIR'), '..', 'piplist', 'build',
                               'cmt_top_lower', wirelist)) as f:
            for l in f:
                tile_type, wire = l.strip().split()

                if tile_type not in wiredata:
                    wiredata[tile_type] = set()

                wiredata[tile_type].add(wire)

    # Load PPIP lists (to exclude them)
    print("Loading PPIP lists...")
    for ppiplist in ppiplists:
        fname = os.path.join(
            os.getenv('FUZDIR'), '..', '071-ppips', 'build', ppiplist)
        with open(fname, 'r') as f:
            for l in f:
                pip_data, pip_type = l.strip().split()

                if pip_type != 'always':
                    continue

                tile_type, dst, src = pip_data.split('.')
                if tile_type not in ppipdata:
                    ppipdata[tile_type] = []

                ppipdata[tile_type].append((src, dst))

    # Load desgin data
    print("Loading design data...")
    with open("design.txt", "r") as f:
        for line in f:
            fields = line.strip().split(",")
            designdata[fields[0]] = fields[1:]

    with open("design_pips.txt", "r") as f:
        for line in f:
            tile, pip, src, dst, pnum, pdir = line.split()

            if not tile.startswith('CMT_TOP'):
                continue

            if 'UPPER_B' in tile:
                continue
            if 'LOWER_T' in tile:
                continue

            pip_prefix, _ = pip.split(".")
            tile_from_pip, tile_type = pip_prefix.split('/')
            assert tile == tile_from_pip
            _, src = src.split("/")
            _, dst = dst.split("/")
            pnum = int(pnum)
            pdir = int(pdir)

            if tile not in tiledata:
                tiledata[tile] = {
                    "type": tile_type,
                    "pips": set(),
                    "srcs": set(),
                    "dsts": set(),
                }

            tiledata[tile]["pips"].add((src, dst))
            tiledata[tile]["srcs"].add(src)
            tiledata[tile]["dsts"].add(dst)

            if pdir == 0:
                tiledata[tile]["srcs"].add(dst)
                tiledata[tile]["dsts"].add(src)

            #if dst.startswith('CMT_TOP_R_LOWER_B_CLK') or \
            #   dst.startswith('CMT_TOP_L_LOWER_B_CLK'):
            #    ignpip.add((src, dst))

    active_wires = {}
    with open("design_wires.txt", "r") as f:
        for l in f:
            tile, wire = l.strip().split('/')

            if tile not in active_wires:
                active_wires[tile] = set()

            active_wires[tile].add(wire)

    tags = {}

    # Populate IN_USE tags
    for tile, (site, in_use) in designdata.items():
        if tile not in tags:
            tags[tile] = {}

        tile_type = tile.rsplit("_", maxsplit=1)[0]
        tags[tile]["IN_USE"] = int(in_use)

    # Populate PIPs
    active_clks = {}
    for tile in tags.keys():
        tile_type = tile.rsplit("_", maxsplit=1)[0]

        in_use = tags[tile]["IN_USE"]

        if not in_use:
            active_pips = []
        else:
            active_pips = tiledata[tile]["pips"]

        for src, dst in pipdata[tile_type]:

            if (src, dst) in ignpip:
                continue
            if (src, dst) in ppipdata[tile_type]:
                continue

            tag = "{}.{}".format(merge_lr_wires(dst), merge_lr_wires(src))
            val = in_use if (src, dst) in active_pips else False

            if not (in_use and not val):
                if tile not in active_clks:
                    active_clks[tile] = set()

                active_clks[tile].add(src)
                tags[tile][tag] = int(val)

        for wire in wiredata[tile_type]:
            if 'CLK' not in wire:
                continue

            if 'CLKOUT' in wire:
                continue

            if 'CLKFB' in wire:
                continue

            if 'REBUF' in wire:
                continue

            wire = merge_lr_wires(wire)

            if tile not in active_wires:
                active_wires[tile] = set()
            segmk.add_tile_tag(
                tile, '{}_ACTIVE'.format(wire), wire in active_wires[tile])

    # Output tags
    for tile, tile_tags in tags.items():
        for t, v in tile_tags.items():
            segmk.add_tile_tag(tile, t, v)

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import math
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray import lut_maker
from prjxray.db import Database

NOT_INCLUDED_TILES = ['LIOI3_SING', 'RIOI3_SING']

SITE_TYPES = ['OLOGICE3', 'ILOGICE3']


def read_site_to_cmt():
    """ Yields clock sources and which CMT they route within. """
    with open(os.path.join(os.getenv('FUZDIR'), 'build',
                           'cmt_regions.csv')) as f:
        for l in f:
            site, cmt = l.strip().split(',')
            yield (site, cmt)


def gen_sites():
    ''' Return dict of ISERDES/OSERDES locations. '''
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    xy_fun = util.create_xy_fun('\S+')

    tiles = grid.tiles()

    for tile_name in sorted(tiles):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        tile_type = gridinfo.tile_type

        tile = {'tile': tile_name, 'tile_type': tile_type, 'ioi_sites': {}}

        for site_name, site_type in gridinfo.sites.items():
            if site_type in SITE_TYPES:
                xy = xy_fun(site_name)
                if xy not in tile['ioi_sites']:
                    tile['ioi_sites'][xy] = {}

                tile['ioi_sites'][xy][site_type] = site_name

        yield tile


class ClockSources(object):
    def __init__(self):
        self.site_to_cmt = dict(read_site_to_cmt())

        self.leaf_gclks = {}
        self.ioclks = {}
        self.rclks = {}
        self.selected_leaf_gclks = {}
        self.lut_maker = lut_maker.LutMaker()

        for cmt in set(self.site_to_cmt.values()):
            self.leaf_gclks[cmt] = []
            self.ioclks[cmt] = []
            self.rclks[cmt] = []

    def init_clocks(self):
        """ Initialize all IOI clock sources. """
        for site, cmt in self.site_to_cmt.items():
            clk = 'clk_' + site
            if 'BUFHCE' in site:
                print(
                    """
            wire {clk};
            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            BUFH bufh_{site}(
                .O({clk})
                );
                """.format(
                        clk=clk,
                        site=site,
                    ))

                self.leaf_gclks[cmt].append(clk)

            if 'BUFIO' in site:
                print(
                    """
            wire {clk};
            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            BUFIO bufio_{site}(
                .O({clk})
                );
                """.format(
                        clk=clk,
                        site=site,
                    ))

                self.ioclks[cmt].append(clk)

            if 'BUFR' in site:
                print(
                    """
            wire {clk};
            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            BUFR bufr_{site}(
                .O({clk})
                );
                """.format(
                        clk=clk,
                        site=site,
                    ))

                self.rclks[cmt].append(clk)

        # Choose 6 leaf_gclks to be used in each CMT.
        for cmt in self.leaf_gclks:
            self.selected_leaf_gclks[cmt] = random.sample(
                self.leaf_gclks[cmt], 6)

    def get_clock(
            self,
            site,
            allow_ioclks,
            allow_rclks,
            allow_fabric=True,
            allow_empty=True):
        cmt = self.site_to_cmt[site]
        choices = []
        if allow_fabric:
            choices.append('lut')

        if allow_empty:
            choices.append('')

        choices.extend(self.selected_leaf_gclks[cmt])
        if allow_ioclks:
            choices.extend(self.ioclks[cmt])

        if allow_rclks:
            choices.extend(self.rclks[cmt])

        clock = random.choice(choices)
        is_lut = False
        if clock == "lut":
            clock = self.lut_maker.get_next_output_net()
            is_lut = True
        return clock, is_lut


def add_port(ports, port, signal):
    ports.append('.{}({})'.format(port, signal))


def run():
    print("module top();")

    clocks = ClockSources()
    clocks.init_clocks()
    """

    ISERDESE2 clock sources:

    CLK/CLKB:
     - Allows LEAF_GCLK, IOCLKS, RCLKS and fabric
     - Dedicated pips

    CLKDIV:
     - No dedicated pips, uses fabric clock in.

    CLKDIVP:
     - Has pips, MIG only, PHASER or fabric.

    OCLK/OCLKB:
     - Allows LEAF_GCLK, IOCLKS, RCLKS and fabric
     - Must match OSERDESE2:CLK/CLKB

     OSERDESE2 clock sources:

     CLKDIV/CLKDIVB:
      - Allows LEAF_GCLK and RCLKS and fabric
      - Dedicated pips

     CLKDIVF/CLKDIVFB:
      - Allows LEAF_GCLK and RCLKS and fabric
      - No explicit port, follows CLKDIV/CLKDIVB?
      """

    output = []
    route_file = open("routes.txt", "w")

    for tile in gen_sites():
        if tile['tile_type'] in NOT_INCLUDED_TILES:
            continue

        for xy in tile['ioi_sites']:
            ilogic_site_type = random.choice([None, 'ISERDESE2', 'IDDR'])
            use_oserdes = random.randint(0, 1)

            ilogic_site = tile['ioi_sites'][xy]['ILOGICE3']
            ologic_site = tile['ioi_sites'][xy]['OLOGICE3']

            if use_oserdes:
                oclk, _ = clocks.get_clock(
                    ologic_site, allow_ioclks=True, allow_rclks=True)

                oclkb = oclk
            else:
                oclk, is_lut = clocks.get_clock(
                    ilogic_site, allow_ioclks=True, allow_rclks=True)

                if random.randint(0, 1):
                    oclkb = oclk
                else:
                    if random.randint(0, 1):
                        oclkb, _ = clocks.get_clock(
                            ilogic_site,
                            allow_ioclks=True,
                            allow_rclks=True,
                            allow_fabric=not is_lut)
                    else:
                        # Explicitly provide IMUX stimulus to resolve IMUX pips
                        oclk = random.randint(0, 1)
                        oclkb = random.randint(0, 1)

            DATA_RATE = random.choice(['DDR', 'SDR'])
            clk, clk_is_lut = clocks.get_clock(
                ilogic_site,
                allow_ioclks=True,
                allow_rclks=True,
                allow_empty=DATA_RATE == 'SDR')

            clkb = clk
            while clkb == clk:
                clkb, clkb_is_lut = clocks.get_clock(
                    ilogic_site,
                    allow_ioclks=True,
                    allow_rclks=True,
                    allow_empty=False)

            imux_available = {
                0: set(("IOI_IMUX20_0", "IOI_IMUX22_0")),
                1: set(("IOI_IMUX20_1", "IOI_IMUX22_1")),
            }

            # Force CLK route through IMUX when connected to a LUT
            if clk_is_lut:
                y = (xy[1] + 1) % 2

                route = random.choice(list(imux_available[y]))
                imux_available[y].remove(route)

                route = "{}/{}".format(tile["tile"], route)
                route_file.write("{} {}\n".format(clk, route))

            # Force CLKB route through IMUX when connected to a LUT
            if clkb_is_lut:
                y = (xy[1] + 1) % 2

                route = random.choice(list(imux_available[y]))
                imux_available[y].remove(route)

                route = "{}/{}".format(tile["tile"], route)
                route_file.write("{} {}\n".format(clkb, route))

            if ilogic_site_type is None:
                pass

            elif ilogic_site_type == 'ISERDESE2':
                INTERFACE_TYPE = random.choice(
                    [
                        'MEMORY',
                        'MEMORY_DDR3',
                        'MEMORY_QDR',
                        'NETWORKING',
                        'OVERSAMPLE',
                    ])
                ports = []

                add_port(ports, 'CLK', clk)
                add_port(ports, 'CLKB', clkb)
                add_port(ports, 'OCLK', oclk)
                add_port(ports, 'OCLKB', oclkb)

                output.append(
                    """
            (* KEEP, DONT_TOUCH, LOC="{site}" *)
            ISERDESE2 #(
                .DATA_RATE({DATA_RATE}),
                .INTERFACE_TYPE({INTERFACE_TYPE}),
                .IS_CLK_INVERTED({IS_CLK_INVERTED}),
                .IS_CLKB_INVERTED({IS_CLKB_INVERTED}),
                .IS_OCLK_INVERTED({IS_OCLK_INVERTED}),
                .IS_OCLKB_INVERTED({IS_OCLKB_INVERTED}),
                .INIT_Q1({INIT_Q1}),
                .INIT_Q2({INIT_Q2}),
                .INIT_Q3({INIT_Q3}),
                .INIT_Q4({INIT_Q4}),
                .SRVAL_Q1({SRVAL_Q1}),
                .SRVAL_Q2({SRVAL_Q2}),
                .SRVAL_Q3({SRVAL_Q3}),
                .SRVAL_Q4({SRVAL_Q4})
            ) iserdes_{site}(
                {ports});""".format(
                        site=ilogic_site,
                        ports=',\n'.join(ports),
                        DATA_RATE=verilog.quote(DATA_RATE),
                        INTERFACE_TYPE=verilog.quote(INTERFACE_TYPE),
                        IS_CLK_INVERTED=random.randint(0, 1),
                        IS_CLKB_INVERTED=random.randint(0, 1),
                        IS_OCLK_INVERTED=random.randint(0, 1),
                        IS_OCLKB_INVERTED=random.randint(0, 1),
                        INIT_Q1=random.randint(0, 1),
                        INIT_Q2=random.randint(0, 1),
                        INIT_Q3=random.randint(0, 1),
                        INIT_Q4=random.randint(0, 1),
                        SRVAL_Q1=random.randint(0, 1),
                        SRVAL_Q2=random.randint(0, 1),
                        SRVAL_Q3=random.randint(0, 1),
                        SRVAL_Q4=random.randint(0, 1),
                    ))
            elif ilogic_site_type == 'IDDR':
                ports = []
                add_port(ports, 'C', clk)
                add_port(ports, 'CB', clkb)

                output.append(
                    """
            (* KEEP, DONT_TOUCH, LOC="{site}" *)
            IDDR_2CLK #(
                .INIT_Q1({INIT_Q1}),
                .INIT_Q2({INIT_Q2}),
                .SRTYPE({SRTYPE})
            ) iserdes_{site}(
                {ports});""".format(
                        site=ilogic_site,
                        ports=',\n'.join(ports),
                        INIT_Q1=random.randint(0, 1),
                        INIT_Q2=random.randint(0, 1),
                        SRTYPE=verilog.quote(random.choice(['ASYNC', 'SYNC'])),
                    ))
            else:
                assert False, ilogic_site_type

            if use_oserdes:
                ports = []

                add_port(
                    ports, 'CLKDIV',
                    clocks.get_clock(
                        ologic_site,
                        allow_ioclks=False,
                        allow_rclks=True,
                    )[0])

                add_port(ports, 'CLK', oclk)

                output.append(
                    """
            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            OSERDESE2 #(
                .IS_CLK_INVERTED({IS_CLK_INVERTED}),
                .DATA_RATE_OQ("SDR"),
                .DATA_RATE_TQ("SDR")
            ) oserdes_{site} (
                {ports});""".format(
                        IS_CLK_INVERTED=random.randint(0, 1),
                        site=ologic_site,
                        ports=',\n'.join(ports),
                    ))

    for s in clocks.lut_maker.create_wires_and_luts():
        print(s)

    for s in output:
        print(s)

    print("endmodule")


if __name__ == '__main__':
    run()
